/*
 * bsp, adapted from GGems III by HWN
 */

#include	<stdio.h>
#include 	<stdlib.h>

#include	"ray.h"
#include	"proto.h"
#include 	"extern.h"

/*
 * Supporting Data Structures
 */

typedef struct BinNode BinNode,
               *BinNodePtr;

struct BinNode {
    vector          min,
                    max;	/* extent of node */
    struct optilist *members;	/* list of enclosed primitives */
    struct BinNode *child[2];	/* pointers to children nodes, if any */

    /*
     * distance to the plane which subdivides the children, children near/far ordering relative to ray origin
     */
    void            (*get_division_info) (Flt * dist,
					  BinNode * *near_child,
					  BinNode * *far_child,
					  struct ray * theray, BinNode * currentNode);
};

typedef struct {
    int             MaxDepth;	/* max allowed depth of the tree */
    BinNode        *root;	/* root of the entire bin tree */
    struct optilist *infinites;
}

BinTree;


/* free node of tree, and it's children */
PRIVATE void
free_node(BinNode * p)
{
    free_optimisation_list(p->members);

    /* pointers to children nodes, if any */
    if (p->child[0])
	free_node(p->child[0]);
    if (p->child[1])
	free_node(p->child[1]);

    free(p);
}

/* get and init node */
PRIVATE BinNode *
get_new_node(void)
{
    BinNode        *p;

    p = xmalloc(sizeof(BinNode), "BSP Node");
    p->members = get_new_optimisation_list();
    p->child[0] = p->child[1] = NULL;	/* pointers to children nodes, if any */

    return p;
}


/* get new root entry */
PRIVATE BinTree *
get_new_BSP_tree(void)
{
    BinTree        *t;

    t = xmalloc(sizeof(BinTree), "BSPTree");
    t->root = get_new_node();
    t->infinites = get_new_optimisation_list();
    return t;
}

/* debugging aid */
PRIVATE void
do_print_BSP_tree(BinNode * r)
{
#ifdef DEBUG
    dprintf("BSP {\n");

    if (r->child[0]) {
	do_print_BSP_tree(r->child[0]);
	if (r->child[1])
	    do_print_BSP_tree(r->child[1]);
    } else {
	print_v("min", r->min);
	print_v("max", r->max);
	print_optimisation_list(r->members);
    }
    dprintf("}\n");
#endif				/* DEBUG */
}

/* and more */
PUBLIC void
print_BSP_tree(object *o)
{
#ifdef DEBUG
    BinTree        *BSPTree;
    struct composite_data *c;

    c = o->data.composite;
    BSPTree = c->optidata;

    print_optimisation_list(BSPTree->infinites);

    do_print_BSP_tree(BSPTree->root);
#endif				/* DEBUG */
}


/*
 * Returns the distance between origin and plane, measured along the input
 * direction. direction is a unit vector.
 *
 * Entry: plane     - subdivision plane of current node origin    - origin of
 * the ray dir - direction of the ray, must be a unit vector
 *
 * Exit: returns the distance between the origin and plane measured along the
 * direction
 *
 * Note: there is a function for each of the three subdivision planes
 */



/*
 * Determines which of the half space of the two children contains pos,
 * return that child as near, the other as far.
 *
 * Entry: currentNode - node currently working on pos      - pos of the ray
 *
 * Exit: near - node whose half plane contains the pos far  - node whose half
 * plane does not contain the pos
 *
 * Note: there is a function for each of the three subdivision planes
 */
PRIVATE void
get_x_info(Flt * dist,
	   BinNode * *near_child,
	   BinNode * *far_child,
	   struct ray *theray, BinNode * currentNode)
{

    /*
     * remember that child[0]->max or child[1]->min is the subdivision plane
     */

    if (currentNode->child[0]->max.x >= theray->pos.x) {
	*near_child = currentNode->child[0];
	*far_child = currentNode->child[1];
    } else {
	*far_child = currentNode->child[0];
	*near_child = currentNode->child[1];
    }

    if (theray->dir.x)
	*dist = ((currentNode->child[0]->max.x - theray->pos.x) / theray->dir.x);
    else
	*dist = INFTY;

}

PRIVATE void
get_y_info(Flt * dist,
	   BinNode * *near_child,
	   BinNode * *far_child,
	   struct ray *theray, BinNode * currentNode)
{
    /*
     * remember that child[0]->max or child[1]->min is the subdivision plane
     */

    if (currentNode->child[0]->max.y >= theray->pos.y) {
	*near_child = currentNode->child[0];
	*far_child = currentNode->child[1];
    } else {
	*far_child = currentNode->child[0];
	*near_child = currentNode->child[1];
    }

    if (theray->dir.y)
	*dist = ((currentNode->child[0]->max.y - theray->pos.y) / theray->dir.y);
    else
	*dist = INFTY;

}

PRIVATE void
get_z_info(Flt * dist,
	   BinNode * *near_child,
	   BinNode * *far_child,
	   struct ray *theray, BinNode * currentNode)
{

    /*
     * remember that child[0]->max or child[1]->min is the subdivision plane
     */

    if (currentNode->child[0]->max.z >= theray->pos.z) {
	*near_child = currentNode->child[0];
	*far_child = currentNode->child[1];
    } else {
	*far_child = currentNode->child[0];
	*near_child = currentNode->child[1];
    }

    if (theray->dir.z)
	*dist = ((currentNode->child[0]->max.z - theray->pos.z) / theray->dir.z);
    else
	*dist = INFTY;

}

/* is obj  associated with this node? */
PRIVATE bool
object_in_node(BinNode * node, object *obj)
{
    if (node->min.x > obj->bbox[1].x || node->max.x < obj->bbox[0].x)
	return FALSE;
    if (node->min.y > obj->bbox[1].y || node->max.y < obj->bbox[0].y)
	return FALSE;
    if (node->min.z > obj->bbox[1].z || node->max.z < obj->bbox[0].z)
	return FALSE;
    return TRUE;
}

/*
 * This routine intersects ray with all of the objects in the objList and
 * returns the closest intersection distance and the interesting object,
 * if there is one.
 */

PRIVATE bool
RayObjIntersect(ibuf * q, struct ray *r, struct optilist *objList, int flags, bool *b)
{

    bool            gotcha;
    int             i;


    gotcha = FALSE;
    for (i = 0; i < objList->len; i++) {
	assert(objList->list[i]);

	/* walk the element list, and intersect each element */
	if (intersect_object(q, objList->list[i], r, flags | USEMAIL, b))
	    gotcha = TRUE;

	if (min_param(q) < r->mint)	/* early exit possible? */
	    break;
    }

    return gotcha;
}

#define PUSH(ch, m, M) stack[stackPtr].node = ch; \
		stack[stackPtr].min = m; \
		stack[stackPtr].max = M; \
		stackPtr++;

#define POP(Node, m, M) stackPtr--; \
	Node = stack[stackPtr].node;  \
	m = stack[stackPtr].min; \
	M = stack[stackPtr].max;


/*
 * Data structure for a simple stack. This is necessary for implementing
 * an efficient linear BSP tree walking.
 */
typedef struct {
    BinNode        *node;
    Flt             min,
                    max;
}

StackElem;

PRIVATE StackElem *stack = NULL;
PRIVATE int     maxstack = 0;

/* trace a shadowray through the BSP */
PRIVATE bool
shadow_intersect(ibuf * wait_queue, BinTree * BSPTree, struct ray *theray, int flags, bool *isinside)
{
    int             stackPtr;
    BinNode        *currentNode,
                   *near_child,
                   *far_child;
    Flt             dist,
                    min,
                    max;
    bool            quick,
                    hit;
    bool            gotcha;
    int             i;
    struct optilist *list;


    quick = !(flags & CHKALL);
    {
	Flt             bhits[2];
	int             n;

	n = intersect_rawbox(BSPTree->root->min, BSPTree->root->max, theray, bhits);
	if (!n)
	    return FALSE;
	else if (n == 1) {
	    min = 0;
	    max = bhits[0];
	} else {
	    min = bhits[0];
	    max = bhits[1];
	}
    }


    /* InitStack(stack); */
    /* after having checked everything, NULL is popped of the stack */

    stackPtr = 0;		/* entry 0 is the return value */
    currentNode = BSPTree->root;

    PUSH(NULL, FALSE, TRUE);
    hit = FALSE;


    while (currentNode != NULL) {

	if (min > theray->mint)
	    break;

	while (currentNode->child[0]) {
	    currentNode->get_division_info(&dist,
					   &near_child, &far_child, theray, currentNode);

	    if ((dist > max) || (dist < 0)) {
		currentNode = near_child;
	    } else if (dist < min) {
		currentNode = far_child;
	    } else {
		PUSH(far_child, dist, max);
		currentNode = near_child;
		max = dist;
	    }
	}

	list = currentNode->members;
	gotcha = FALSE;
	for (i = 0; i < list->len; i++) {
	    assert(list->list[i]);

	    /* walk the element list, and intersect each element */
	    if (intersect_object(wait_queue, list->list[i], theray, flags, isinside))
		gotcha = TRUE;

	    if (min_param(wait_queue) < theray->mint)	/* early exit possible? */
		break;
	}

	if (gotcha) {
	    hit = TRUE;
	    /* shadow ray */
	    if (min_param(wait_queue) < theray->mint && (flags & !CHKALL)) {
		break;
	    }
	    if (quick && min_param(wait_queue) > min && min_param(wait_queue) < max) {
		/* only exit if intersection point is in the current node */
		break;
	    }
	}
	POP(currentNode, min, max);
    }

    return hit;
}

/* trace a normal ray */
PRIVATE bool
normal_intersect(ibuf * wait_queue, BinTree * BSPTree, struct ray *theray, int flags, bool *isinside)
{
    bool            gotcha;
    int             i;
    int             stackPtr;
    BinNode        *currentNode,
                   *near_child,
                   *far_child;
    Flt             dist,
                    min,
                    max;
    bool            quick,
                    hit;
    struct optilist *list;

    quick = !(flags & CHKALL);

    {
	Flt             bhits[2];
	int             n;

	n = intersect_rawbox(BSPTree->root->min, BSPTree->root->max, theray, bhits);
	if (!n)
	    return FALSE;
	else if (n == 1) {
	    min = 0;
	    max = bhits[0];
	} else {
	    min = bhits[0];
	    max = bhits[1];
	}
    }


    /* after having checked everything, NULL is popped of the stack */
    stackPtr = 0;
    currentNode = BSPTree->root;

    PUSH(NULL, FALSE, TRUE);

    hit = FALSE;
    while (currentNode != NULL) {

	while (currentNode->child[0]) {
	    currentNode->get_division_info(&dist,
					   &near_child, &far_child, theray, currentNode);

	    if ((dist > max) || (dist < 0)) {
		currentNode = near_child;
	    } else if (dist < min) {
		currentNode = far_child;
	    } else {
		PUSH(far_child, dist, max);
		currentNode = near_child;
		max = dist;
	    }
	}

	gotcha = FALSE;
	list = currentNode->members;
	for (i = 0; i < list->len; i++) {
	    assert(list->list[i]);

	    /* walk the element list, and intersect each element */
	    if (intersect_object(wait_queue, list->list[i], theray, flags, isinside))
		gotcha = TRUE;
	}

	if (gotcha) {
	    hit = TRUE;
	    if (quick && min_param(wait_queue) > min && min_param(wait_queue) < max) {
		/* only exit if intersection point is in the current node */
		break;
	    }
	}
	POP(currentNode, min, max);
    }

    return hit;
}

/* Tracing entry point. This needs to be really fast! */
PRIVATE bool
raytreeintersect(ibuf * globalq, object *o, struct ray *theray, int flags, bool *isinside)
{
    BinTree        *BSPTree;
    ibuf           *wait_queue;
    bool            hit,
                    useglobal;

    BSPTree = o->data.composite->optidata;
    flags |= USEMAIL | NO_CHKBOUNDS;

    /* try avoid allocating a new intersection buffer */
    if (flags & CHKALL || ibuf_empty(globalq)) {
	wait_queue = globalq;
	useglobal = TRUE;
    } else {
	useglobal = FALSE;
	wait_queue = get_new_ibuf();
    }

    if (theray->type == R_SHADOW)
	hit = shadow_intersect(wait_queue, BSPTree, theray, flags, isinside);
    else
	hit = normal_intersect(wait_queue, BSPTree, theray, flags, isinside);

    if (!useglobal) {
	add_to_ibuf(globalq, flags, &(min_entry(wait_queue)));
	free_ibuf(wait_queue);
    }
    hit = RayObjIntersect(globalq, theray, BSPTree->infinites, flags, isinside) | hit;
    return hit;
}



/*
 * Builds the BSP tree by subdividing along the center of x, y, or z
 * bounds one each time this function is called.
 */
PRIVATE int
subdivide(BinNode * node,
	  int depth,		/* current tree depth */
	  int *maxdepth,	/* max reached depth */
	  int MaxListLength,	/* the specified max allowed list length */
	  int axis)
{				/* current subdivision plane, 1-x, 2-y, 3-z */
    int             i,
                    objcount,
                    leafs,
                    threshold,
                    nextAxis;
    BinNode        *childp;
    Flt             minfactor,
                    maxfactor,
                    m;

    clean_optimisation_list(node->members);
    if (*maxdepth < depth)
	*maxdepth = depth;

    if ((node->members->len > MaxListLength)) {
	for (i = 0; i < 2; i++) {
	    m = 0.5;		/* always divide in two equal parts */
	    minfactor = (m * i);
	    maxfactor = ((1 - m) * (1 - i) + i);

	    node->child[i] = childp = get_new_node();

	    childp->min = node->min;
	    childp->max = node->max;

	    switch (axis) {
	    case 1:
		node->get_division_info = get_x_info;

		/* current subdivision plane is x */
		childp->min.x =
		    node->min.x + minfactor * (node->max.x - node->min.x);
		childp->max.x =
		    node->min.x + maxfactor * (node->max.x - node->min.x);

		/* child subdivision plane will be y */
		nextAxis = 2;
		break;
	    case 2:

		node->get_division_info = get_y_info;

		/* current subdivision plane is y */
		childp->min.y =
		    node->min.y + minfactor * (node->max.y - node->min.y);
		childp->max.y =
		    node->min.y + maxfactor * (node->max.y - node->min.y);

		/* child subdivision plane will be z */
		nextAxis = 3;
		break;
	    case 3:
		node->get_division_info = get_z_info;
		/* current subdivision plane is z */
		childp->min.z =
		    node->min.z + minfactor * (node->max.z - node->min.z);
		childp->max.z =
		    node->min.z + maxfactor * (node->max.z - node->min.z);

		/* child subdivision plane will be x */
		nextAxis = 1;
		break;
	    }

	    for (objcount = 0; objcount < node->members->len; objcount++) {
		if (object_in_node(childp, node->members->list[objcount])) {

		    /* put stuff in children */
		    add_to_optilist(childp->members, node->members->list[objcount]);
		}
	    }
	}

	threshold = (node->members->len * 9) / 10;

	if (node->child[0]->members->len >= threshold && node->child[1]->members->len >= threshold) {

	    /* this step wasn't really useful.. */
	    free_node(node->child[0]);
	    free_node(node->child[1]);
	    node->child[0] = NULL;
	    node->child[1] = NULL;

	    /* Try again. Beware. this is recursion */
	    leafs = subdivide(node, depth + 1, maxdepth, (MaxListLength * 5) / 4 + 1, nextAxis);
	} else {
	    free_optimisation_list(node->members);
	    node->members = NULL;
	    leafs = subdivide(node->child[0], depth + 1, maxdepth, MaxListLength, nextAxis);
	    leafs += subdivide(node->child[1], depth + 1, maxdepth, MaxListLength, nextAxis);
	}


    } else
	leafs = 1;

    return leafs;
}

/*
 * Initialize and start the building of BSPTree.
 */
PUBLIC void
init_BSP(object *o)
{
    BinTree        *BSPTree;
    struct composite_data *c;
    struct optilist *tmplist;
    int             max,
                    leafs,
                    leafs2,
                    i;

    c = o->data.composite;
    c->optidata = BSPTree = get_new_BSP_tree();

    tmplist = get_new_optimisation_list();
    build_list(o, tmplist);
    for (i = 0; i < tmplist->len; i++) {
	if (tmplist->list[i]->flags & O_INFINITE)
	    add_to_optilist(BSPTree->infinites, tmplist->list[i]);
	else
	    add_to_optilist(BSPTree->root->members, tmplist->list[i]);
    }
    free_optimisation_list(tmplist);
    clean_optimisation_list(BSPTree->infinites);


    c->all_intersections_method = raytreeintersect;
    c->print_optimisation = print_BSP_tree;

    /* Start building the BSPTree by subdividing along the x axis first */
    BSPTree->root->min = BSPTree->root->members->bmin;
    BSPTree->root->max = BSPTree->root->members->bmax;

    max = 0;
    leafs = subdivide(BSPTree->root, 0, &max, bunching, 1);

    printf("BSP Depth %d, leafs %d\n", max, leafs);
    max += 2;
    if (max > maxstack) {
	maxstack = max;
	stack = xrealloc(stack, sizeof(StackElem) * maxstack * 3, "BSP Stack");
    }
}
