/*
 * conic.c -- handle conic type objects
 *
 * (c) 1993, 1994 by Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 * inspired on Jarke J. van Wijk's thesis on procedural primitives
 */

#include "ray.h"
#include "proto.h"
#include "extern.h"

extern struct methods my_methods;

PRIVATE void
free_conic(object *o)
{
    free_object(o->data.conic->shape);
    free((void *) o->data.conic);

    o->type = NOSHAPE;
}

PRIVATE bool
inside_conic(object *o, vector i)
{
    struct conic_data *c = o->data.conic;

    vector          d,
                    loc;
    Flt             t;

    if (i.z > c->height || i.z < 0)
	return FALSE;

    vsub(d, i, c->apex);

    if (ISZERO(d.z))
	return FALSE;

    t = -c->apex.z / d.z;
    svproduct(loc, t, d);
    vadd(loc, c->apex, loc);
    return inside_object(c->shape, loc);
}

PRIVATE void
print_conic(object *o)
{
#ifdef DEBUG
    struct conic_data *p = o->data.conic;

    dprintf("cap: lo %d, hi %d.", p->locap, p->hicap);
    dprintf("Height %f\n", p->height);
    print_v("apex", p->apex);
    dprintf("Base object:\n");
    print_object(p->shape);
#endif				/* DEBUG */
}


PRIVATE bool
all_conic_intersections(ibuf * globalq, object *o, struct ray *inray, int flags, bool *isinside)
{
    dqueue          q_ent;
    struct conic_data *c;
    struct ray      projray;	/* ray projected onto XY plane */
    int             rechit;
    Flt             t;
    bool            inside;
    struct ray      localray;

    my_methods.test++;
    rechit = 0;

    q_ent.obj = o;
    c = o->data.conic;


    /* project ray from apex onto xy plane. */
    {
	vector          p1,
	                p2,
	                dir;
	Flt             dz;

	p1 = inray->pos;
	dz = p1.z - c->apex.z;
	if (ISZERO(dz)) {
	    svproduct(p1, tolerance / 10, inray->dir);
	    vadd(p1, inray->pos, p1);
	    dz = p1.z - c->apex.z;
	}
	if (ISZERO(dz)) {
	    return 0;
	    *isinside = o->flags & O_INVERTED;
	}
	t = -c->apex.z / dz;
	vsub(dir, p1, c->apex);
	svproduct(dir, t, dir);
	vadd(p1, dir, c->apex);

	vadd(p2, inray->pos, inray->dir);
	dz = p2.z - c->apex.z;

	if (ISZERO(dz)) {
	    svproduct(p2, -tolerance / 10, inray->dir);
	    vadd(p2, inray->pos, p2);
	    dz = p2.z - c->apex.z;
	}
	if (ISZERO(dz)) {
	    return 0;
	    *isinside = o->flags & O_INVERTED;
	}
	t = -c->apex.z / dz;
	vsub(dir, p2, c->apex);
	svproduct(dir, t, dir);
	vadd(p2, dir, c->apex);

	localray = *inray;
	localray.pos = p1;
	vsub(localray.dir, p2, p1);
    }


    if (ISZERO(inray->dir.x) &&
	ISZERO(inray->dir.y)) {	/* vertical? */
	Flt             t1,
	                t2;

	/* inside ? */
	if (inside_object(c->shape, projray.pos)) {

	    /* then check intersections with horiz. planes */
	    if (c->locap)
		t1 = -inray->pos.z / inray->dir.z;
	    else
		t1 = -INFTY;

	    if (c->hicap)
		t2 = (c->height - inray->pos.z) / inray->dir.z;
	    else
		t2 = -INFTY;

	    if (t1 > tolerance)
		rechit++;
	    if (t2 > tolerance)
		rechit++;

	    inside = (*isinside) = (!!(o->flags & O_INVERTED)) ^ (rechit % 2);
	    q_ent.obj2 = NULL;

	    if (t1 > tolerance) {
		q_ent.entering = (inside = !inside);
		q_ent.t = t1;
		add_to_ibuf(globalq, flags, &q_ent);
	    }
	    if (t2 > tolerance) {
		q_ent.entering = (inside = !inside);
		q_ent.t = t2;
		add_to_ibuf(globalq, flags, &q_ent);
	    }
	} else {
	    /*
	     * ray from apex which is not inside. This will not hit the shape
	     */
	    *isinside = !(o->flags & O_INVERTED);
	    return 0;
	}
    } else {
	ibuf           *q = get_new_ibuf();

	dqueue
	    * qp;
	bool            gotbot,
	                xinside;
	Flt             zcoord,
	                ztop,
	                zbot;



	/* Get the intersections of the generating shape */
	intersect_object(q, c->shape, &localray, CHKALL | CHKINSIDE, &inside);

	if (inray->dir.z >= 0) {
	    zbot = 0.0;
	    ztop = c->height;
	} else {

	    /* convert the problem */
	    zbot = -c->height;
	    ztop = 0.0;
	    inray->dir.z = -inray->dir.z;
	    inray->pos.z = -inray->pos.z;
	}

	/* inside conic */
	*isinside = xinside = (inray->pos.z >= zbot && inray->pos.z <= ztop && inside) ^ (!!(o->flags & O_INVERTED));

	gotbot = FALSE;

	first_entry(q);
	for (qp = next_entry(q); qp != NULL; qp = next_entry(q)) {
	    inside = !inside;
	    zcoord = qp->t * inray->dir.z + inray->pos.z;

	    /* check bottom intersection */
	    if (zcoord >= zbot && !gotbot) {

		if (!inside && c->locap) {
		    t = (zbot - inray->pos.z) / inray->dir.z;
		    if (t > tolerance) {
			q_ent.t = t;
			rechit++;
			q_ent.entering = (xinside = !xinside);
			add_to_ibuf(globalq, flags, &q_ent);
			if (!(flags & CHKALL))
			    break;
		    }
		}
		gotbot = TRUE;
	    }
	    /* check top intersection */
	    if (zcoord >= ztop) {
		if (c->hicap && !inside) {
		    t = (ztop - inray->pos.z) / inray->dir.z;
		    if (t > tolerance) {
			q_ent.t = t;
			q_ent.entering = (xinside = !xinside);
			rechit++;
			add_to_ibuf(globalq, flags, &q_ent);
			if (!(flags & CHKALL))
			    break;
		    }
		}
		break;
	    }
	    /*
	     * we've passed the other tests. This is an intersection with the standing part of the conic
	     */
	    if (zcoord >= zbot) {
		q_ent.t = qp->t;
		q_ent.obj2 = qp->obj;
		q_ent.entering = (xinside = !xinside);

		rechit++;
		add_to_ibuf(globalq, flags, &q_ent);
		if (!(flags & CHKALL))
		    break;
	    }
	}

	free_ibuf(q);
    }
    if (rechit)
	my_methods.hit++;

    return rechit;
}

/*
 * returns the normal to the conic
 */
PRIVATE vector
conic_normal(struct intersect *i, vector loc)
{
    struct conic_data *c;

    vector
                    n;

    c = i->q_ent.obj->data.conic;

    if (ISZERO(loc.z)) {
	setvector(n, 0, 0, -1);
    } else if (ISZERO(loc.z - c->height)) {
	setvector(n, 0, 0, 1);
    } else {
	vector          raakvect,
	                toapex,
	                d;
	struct intersect i2;
	Flt             t;

	vsub(d, loc, c->apex);
	assert(!ISZERO(d.z));

	t = -c->apex.z / d.z;
	svproduct(loc, t, d);
	vadd(loc, c->apex, loc);
	assert(ISZERO(loc.z));

	i2 = *i;
	i2.q_ent.obj = i->q_ent.obj2;
	n = object_normal(&i2, loc);
	print_v("flat", n);
	vsub(toapex, c->apex, loc);
	vcross(raakvect, toapex, n);
	vcross(n, raakvect, toapex);
	norm(n, n);

    }
    print_v("n", n);

    return n;
}

/* initialize a conic_data struct */
PRIVATE void
init_conic(struct conic_data *t)
{
    t->shape = NULL;
    t->locap = t->hicap = TRUE;
    t->height = 0;
}

/* alloc and init conic */
PRIVATE struct conic_data *
get_new_conic(void)
{
    struct conic_data *p;
    p = ALLOC(struct conic_data);


    init_conic(p);
    return p;
}

/* copy the shape data */
PRIVATE void
copy_conic(object *dst, object *src)
{
    assert(dst != NULL && src != NULL);

    if (dst->type != CONIC)
	dst->data.conic = get_new_conic();
    *dst->data.conic = *src->data.conic;
    dst->type = src->type;

    if (src->data.conic->shape != NULL) {
	dst->data.conic->shape = get_new_object();
	dst->data.conic->shape->daddy = dst;
    }
    copy_object(dst->data.conic->shape, src->data.conic->shape);
}

PRIVATE void
precompute_conic(object *o)
{
    object         *child;
    struct conic_data *x;
    vector          p1,
                    p2;

    my_methods.howmuch++;
    global_stats.nonprims++;

    x = o->data.conic;
    child = x->shape;
    precompute_object(child);
    o->flags |= child->flags & O_INFINITE;

    p1 = child->bbox[0];
    p1.z = 0.0;
    p2 = child->bbox[1];
    p2.z = x->height;

    do_box_bound(&o->bbox[0], &o->bbox[1], *o->trans, p1, p2);
}


PRIVATE struct methods my_methods =
{
    all_conic_intersections,
    conic_normal,
    copy_conic,
    inside_conic,
    generic_rotate_object,
    generic_translate_object,
    generic_scale_object,
    free_conic,
    print_conic,
    precompute_conic,
    "conic",
};


/* alloc/init */
PUBLIC object  *
get_new_conic_object(void)
{
    object         *o;

    o = get_new_object();

    o->data.conic = get_new_conic();
    o->methods = &my_methods;
    o->type = CONIC;
    o->flags = O_CLOSED | O_FINITE_INSIDE;
    return o;
}

/* create a cone running from <p1> to  <p2> */
PUBLIC object  *
make_cone_object(vector p1, vector p2, Flt rad1, Flt rad2)
{
    vector          dir;
    struct conic_data *x;
    object         *o;
    Flt             theta,
                    phi;
    vector          rot;
    matrix          rotmat;

    if (ISZERO(rad1 - rad2))
	return make_cylinder_object(p1, p2, rad1);

    o = get_new_conic_object();

    x = o->data.conic;

    vsub(dir, p2, p1);
    x->height = veclen(dir);
    x->shape = get_new_sphere_object();
    x->shape->data.sphere->radius = rad1;

    /* express p2 - p1 in cylinder coordinates */
    theta = acos(dir.z / veclen(dir));
    phi = safe_arctangent(dir.y, dir.x);

    /* rotate accordingly */
    setvector(rot, 0, theta, phi);
    rotate_matrix(rotmat, rot);
    rotate_object(o, rotmat);
    translate_object(o, p1);

    return o;
}

/* eof */
