/*
 * grids.c -- handle grids and gridtrees
 *
 * (c) 1994 by Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include	"ray.h"
#include	"proto.h"
#include 	"extern.h"

/*
 * alloc a new grid_tree entry, and construct a permutation on the global
 * grid's indices
 */
PRIVATE grid_tree *
get_new_grid(void)
{
    grid_tree      *g;
    bool           *tmp_indices;
    int             i,
                    j;

    g = ALLOC(grid_tree);

    g->reflect_grid = g->refract_grid = NULL;
    g->scache = xmalloc(sizeof(object *) * light_list_len, "shadow cache");
    g->indices = xmalloc(sizeof(int) * grid_len, "grid indices");
    g->shadowindices = xmalloc(sizeof(int) * grid_len, "shadow indices");

    for (i = 0; i < light_list_len; i++)
	g->scache[i] = NULL;

    g->recursion = g->last = g->shadowlast = 0;

    tmp_indices = xmalloc(sizeof(bool) * grid_len, "temp grid indices");

    for (i = 0; i < grid_len; i++)
	tmp_indices[i] = FALSE;

    for (i = 0; i < grid_len; i++) {
	do {
	    j = rand() % grid_len;	/* guess an index */
	} while (tmp_indices[j]);	/* have we used it earlier? */
	g->indices[i] = j;
	tmp_indices[j] = TRUE;
    }

    for (i = 0; i < grid_len; i++)
	tmp_indices[i] = FALSE;

    for (i = 0; i < grid_len; i++) {
	do {
	    j = rand() % grid_len;	/* guess an index */
	} while (tmp_indices[j]);	/* have we used it earlier? */
	g->shadowindices[i] = j;
	tmp_indices[j] = TRUE;
    }

    free(tmp_indices);
    return g;
};

/* sample (u,v) from the grid entry g for a ray of type raytype */
PUBLIC void
sample_grid(Flt * u, Flt * v, Flt * f, grid_tree * g, int raytype)
{
    Flt            *dp;
    grid_tree      *child;
    int             index;

    switch (raytype) {
    case R_SHADOW:
	index = g->shadowindices[g->shadowlast];
	g->shadowlast = (g->shadowlast + 1) % grid_len;
	break;
    case R_REFLECT:
	if (!g->reflect_grid)
	    g->reflect_grid = get_new_grid();
	child = g->reflect_grid;
	assert(child != NULL);
	index = child->indices[child->last];
	child->last = (child->last + 1) % grid_len;
	break;
    case R_REFRACT:
	if (!g->refract_grid)
	    g->refract_grid = get_new_grid();
	child = g->refract_grid;
	assert(child != NULL);
	index = child->indices[child->last];
	child->last = (child->last + 1) % grid_len;
	break;
    case R_EYE:
	if (!the_grid_tree)
	    the_grid_tree = get_new_grid();

	/* access global tree directly */
	index = the_grid_tree->indices[the_grid_tree->last];
	the_grid_tree->last = (the_grid_tree->last + 1) % grid_len;
	break;
    default:
	assert(FALSE);
    }
    dp = the_grid + 2 * index;

    *u = *dp++ + jitter_width_x * rand1();
    *v = *dp++ + jitter_width_y * rand1();
    *f = filter[index];
}

/* print out info from tree node g, and recurse */
PRIVATE void
print_grid_tree(grid_tree * g)
{
    int             i;

    dprintf("\nindices: ");
    for (i = 0; i < grid_len; i++) {
	dprintf("%d ", g->indices[i]);
    }
    dprintf(" shadow indices: ");
    for (i = 0; i < grid_len; i++) {
	dprintf("%d ", g->shadowindices[i]);
    }
    dprintf("\n");

    if (g->refract_grid != NULL) {
	dprintf("refract grid {");
	print_grid_tree(g->refract_grid);
	dprintf("}\n");
    }
    if (g->reflect_grid != NULL) {
	dprintf("reflect grid {");
	print_grid_tree(g->reflect_grid);
	dprintf("}\n");
    }
}

/* print all there is to know about grids */
PUBLIC void
print_grid(void)
{
    int             i;
    Flt            *dp,
                    u,
                    v;

    dprintf("%d entries, maxjitter (%f,%f) \n2d grid {\n", grid_len + 1, jitter_width_x, jitter_width_y);
    dp = the_grid;
    for (i = 0; i < grid_len; i++) {
	u = *dp++;
	v = *dp++;
	dprintf("(%f, %f)", u, v);
    }
    dprintf("}\nfilter {\n");
    dp = filter;
    for (i = 0; i < grid_len; i++) {
	dprintf("%f ", *dp++);
    }
    dprintf("}\n");

    print_grid_tree(the_grid_tree);
}

/* recursively build the grid */
PRIVATE void
build_grid(grid_tree * g)
{
    if (g->recursion > max_trace_level)
	return;

    g->refract_grid = get_new_grid();
    g->refract_grid->recursion = g->recursion + 1;

    build_grid(g->refract_grid);
    g->reflect_grid = get_new_grid();
    g->reflect_grid->recursion = g->recursion + 1;
    build_grid(g->reflect_grid);
}

/* init children of g */
PUBLIC void
add_grid(grid_tree * g)
{
    if (g->recursion > max_trace_level)
	return;

    if (!g->refract_grid) {
	g->refract_grid = get_new_grid();
	g->refract_grid->recursion = g->recursion + 1;
    }
    if (!g->reflect_grid) {
	g->reflect_grid = get_new_grid();
	g->reflect_grid->recursion = g->recursion + 1;
    }
}

/* init the global grid */
PUBLIC void
init_global_grid(void)
{
    int             gridx,
                    gridy,
                    x,
                    i,
                    y;
    Flt            *fp,
                   *gp,
                    xcenter,
                    xentry,
                    yentry,
                    ycenter;

    /* try to make a square grid */
    gridx = floor(sqrt(tries));
    gridy = tries / gridx;
    grid_len = gridx * gridy;

    gp = the_grid = xmalloc(2 * grid_len * sizeof(Flt), "grid");
    fp = filter = xmalloc(grid_len * sizeof(Flt), "filter");

    /* centers of a {0,1, .. ,  n-1} interval */
    xcenter = ((Flt) gridx - 1) / ((Flt) 2 * gridx);
    ycenter = ((Flt) gridy - 1) / ((Flt) 2 * gridy);

    if (tries == 1) {
	jitter_width_x = 0;
	jitter_width_y = 0;
    } else {
	jitter_width_x = 1 / ((Flt) gridx);
	jitter_width_y = 1 / ((Flt) gridy);
    }

    for (x = 0; x < gridx; x++)
	for (y = 0; y < gridy; y++) {
	    xentry = 2 * ((Flt) x / (Flt) gridx - xcenter);
	    yentry = 2 * ((Flt) y / (Flt) gridy - ycenter);

	    switch (filter_type) {
	    case GAUSS_FILTER:
		*fp++ = exp(-(sqr(xentry) + sqr(yentry)));
		break;
	    default:
		*fp++ = 1;
		break;
	    }

	    *gp++ = xentry;
	    *gp++ = yentry;
	}


    /* normalize */


    for (i = 0; i < grid_len; i++)
	filterconst += filter[i];
    filterconst = grid_len / filterconst;
    for (i = 0; i < grid_len; i++)
	filter[i] *= filterconst;

    the_grid_tree = get_new_grid();
    /*    build_grid(the_grid_tree);*/
}
