/*
 * ibuf.c -- intersection buffers.
 *
 * (c) 1994 by Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include	"ray.h"
#include	"proto.h"
#include 	"extern.h"

/* make it ready for reuse */
PRIVATE void
reinit_ibuf(ibuf * ib)
{
    ib->len = 0;
    ib->minq.obj = NULL;
    ib->minq.obj2 = NULL;
    ib->minq.t = INFTY;
    ib->free = TRUE;
    ib->sorted = FALSE;
}

/* first time init */
PRIVATE ibuf   *
get_init_ibuf_entry(void)
{
    ibuf           *ib;

    ib = xmalloc(sizeof(ibuf), "intersect buf");
    ib->max = 0;
    ib->list = NULL;
    reinit_ibuf(ib);
    return ib;
}

/*
   get a unused intersection buffer, alloc more if needed, and return
   a pointer to it
 */
PRIVATE ibuf  **ibuffers = NULL;
PRIVATE int     maxibuf = 0;

PUBLIC ibuf    *
get_new_ibuf(void)
{
    int             i,
                    j;


    /* linear search for a new one */
    for (i = 0; i < maxibuf; i++)
	if (ibuffers[i]->free) {
	    ibuffers[i]->free = FALSE;
	    return ibuffers[i];
	}
    /* not found, alloc more */
    maxibuf = maxibuf * 2 + 1;
    ibuffers = xrealloc(ibuffers, maxibuf * sizeof(ibuf *), "intersection buffer");


#ifdef DEBUG
    if (maxibuf > 100)
	errormsg("Too much ibufs in use");
#endif

    for (j = i; j < maxibuf; j++) {
	ibuffers[j] = get_init_ibuf_entry();
    }

    ibuffers[i]->free = FALSE;

    return ibuffers[i];
}

/* notify that "ib" isn't used anymore */
PUBLIC void
free_ibuf(ibuf * ib)
{
    assert(!(ib->free));
    reinit_ibuf(ib);
}

/* add intersection queue */
PUBLIC void
add_to_ibuf(ibuf * ib, int flags, dqueue * q)
{
    assert(!ib->free);
    if (flags & CHKALL) {
	if (ib->len >= ib->max) {
	    ib->max = ib->max * 2 + 1;
	    ib->list = xrealloc(ib->list, ib->max * sizeof(dqueue), "intersection list");
	}
	ib->list[ib->len++] = *q;
	ib->sorted = FALSE;
    }
    /* exponential falloff of light in transparent media */
    if (flags & NO_TRANSUPDATE) {
	if (q->obj->flags & O_TRANSPARENT)
	    return;
    }
    if (q->t < ib->minq.t)
	ib->minq = *q;
}

/* needed for sorting */
PRIVATE int
q_compare(const void *q1, const void *q2)
{
    if (((dqueue *) q1)->t < ((dqueue *) q2)->t)
	return -1;
    if (((dqueue *) q1)->t == ((dqueue *) q2)->t)
	return 0;
    return 1;
}

/* sort intersection buffer */
PUBLIC void
sort_ibuf(ibuf * ib)
{
    if (!ib->sorted) {
	qsort(ib->list, ib->len, sizeof(dqueue), q_compare);
	ib->sorted = TRUE;
    }
    ib->idx = 0;
}


/* print intersection buffer */
PUBLIC void
print_ibuf(ibuf * ib)
{
#ifdef DEBUG
    int             i;

    if (ib->len) {
	for (i = 0; i < ib->len; i++)
	    dprintf("%f ", ib->list[i].t);
	dprintf("\n");
    } else if (ib->minq.obj) {
	dprintf("%f\n", ib->minq.t);
    }
#endif				/* DEBUG */
}
