/*
 * intersect.c -- main intersection routines
 *
 * (c) 1993, 1994 by Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>
 *
 * This file is based on work by  George Kyriazis (1988).
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * snelheids winst: bijhouden of een stilstaand object al getraced is.
 */

#include	<math.h>
#include	"ray.h"
#include	"proto.h"
#include 	"extern.h"

/*
 * Shadow check.  Check if the ray from the intersection point to the
 * light, intersection any objects inbetween.  If it does, we have a
 * shadow. One improvement that can be made is to check the transparencies
 * of all the intersecting objects and make the shadow have a different
 * intensity depending on the transparency of the objects
 */

PUBLIC bool
shadow_test(ibuf * queue, struct ray r, Flt ldistance, int index, grid_tree * g)
{

    object         *op;
    bool            hit,
                    isinside;
    object         *light_object = light_list[index];

    /* not interested in what's happening behind the lightsource */
    r.maxt = ldistance + 10 * tolerance;
    r.mint = ldistance;


    light_object->methods->test++;

    /* do caching ? */
    if (g->scache[index] != NULL) {
	object         *last;

	global_stats.cachetest++;

	/* avoid intersecting CSG parts; stop if we find a composite */
	for (op = g->scache[index]; op->type != COMPOSITE; op = op->daddy)
	    last = op;

	if (intersect_object(queue, last, &r, FALSE, &isinside) && min_param(queue) < ldistance) {
	    global_stats.cachehit++;
	    return TRUE;
	} else {
	    global_stats.cachemiss++;
	    g->scache[index] = NULL;
	}
    }
    line_stats.shadowtest++;
    line_stats.raytest++;

    /* check whole scene */
    if (intersect_object(queue, thescene, &r, 0, &isinside)) {

	/* found intersection */
	line_stats.rayintersections++;

	if (min_param(queue) >= ldistance) {	/* there is nothing blocking the object */
	    g->scache[index] = NULL;
	    hit = FALSE;
	} else {
	    light_object->methods->hit++;
	    g->scache[index] = min_entry(queue).obj;
	    hit = TRUE;
	}

    } else {
	hit = FALSE;		/* no intersections found */

	/* clear cache */
	g->scache[index] = NULL;
    }

    return hit;
}


/*
 * intersect(): intersect everything in the whole scene, and then fill in
 * the intersection details.
 */
PUBLIC struct intersect
intersect(struct ray r)
{
    struct intersect i;
    struct ray      transray;
    object         *op;

    vector          moved;

    line_stats.raytest++;
    switch (r.type) {
    case R_EYE:
	line_stats.eyerays++;
	break;
    case R_REFLECT:
	line_stats.reflected++;
	break;
    case R_REFRACT:
	line_stats.refracted++;
	break;
    default:
	assert(FALSE);
    }


    /* intersect everything in scene */
    r.maxt = INFTY;
    r.mint = 0;


    {
	ibuf           *q;
	bool            isinside;

	q = get_new_ibuf();


	intersect_object(q, thescene, &r, 0, &isinside);
	i.q_ent = min_entry(q);

	free_ibuf(q);


	op = i.q_ent.obj;
    }

    if (op == NULL)
	return i;		/* nothing was found */

    line_stats.rayintersections++;
    setvector(moved, 0.0, 0.0, 0.0);

    if (op->speed) {
	/* What was the translated ray? */
	svproduct(moved, -Time, *op->speed);
    } else
	setvector(moved, 0, 0, 0);

    /*
     * determine the intersection point.
     *
     * i.x = t*transray.dir + transray.pos + obj->speed * Time = t*r.dir + r.pos - speed*Time + speed*Time = t*r.dir + r.pos
     *
     */

    svproduct(i.x, i.q_ent.t, r.dir);
    vadd(i.x, i.x, r.pos);
    vadd(i.objectx, i.x, moved);

    /*
     * determine normal at intersection point
     */
    i.n = i.objectn = object_normal(&i, i.objectx);

    i.r = r;

    return i;
}
