/*
 * lights.c -- standard routines for lights
 *
 * (c) by Han-Wen Nienhuys, 1993 <hanwen@stack.urc.tue.nl>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ray.h"
#include "proto.h"
#include "extern.h"

extern struct methods my_methods;

PRIVATE bool
all_light_intersections(ibuf * q, object *o, struct ray *r, int flags, bool *isinside)
{
    return FALSE;
}


PRIVATE vector
light_normal(struct intersect *i, vector loc)
{
    assert(FALSE);
}

PRIVATE bool
inside_light(object *o, vector i)
{
    assert(FALSE);
}

PRIVATE void
init_light(struct light_data *p)
{
    setvector(p->org, 0, 0, 0);
    setvector(p->direction, 1, 0, 0);
    p->tightness = 0;
    p->type = L_NORMAL;
    setcolor(p->lcolor, 1, 1, 1);
    p->attenuation_exp = 0.0;
    p->radius = 0.0;
}

PRIVATE struct light_data *
get_new_light(void)
{
    struct light_data *p;

    p = ALLOC(struct light_data);

    init_light(p);

    return p;
}

PRIVATE void
scale_light(object *o, vector s)
{
    struct light_data *l = o->data.light;
    vector          inv;

    vcproduct(l->org, l->org, s);
    inv.x = 1 / s.x;
    inv.y = 1 / s.y;
    inv.z = 1 / s.z;
    vcproduct(l->direction, l->direction, inv);
    norm(l->direction, l->direction);
}

PRIVATE void
rotate_light(object *o, matrix rotmat)
{
    struct light_data *l = o->data.light;

    rotate_vector(&l->org, rotmat);
    rotate_vector(&l->direction, rotmat);
}

PRIVATE void
translate_light(object *o, vector t)
{
    struct light_data *l = o->data.light;

    vadd(l->org, l->org, t);
}

PRIVATE void
copy_light(object *dst, object *src)
{
    assert(dst != NULL && src != NULL);

    if (dst->type != LIGHTSOURCE)
	dst->data.light = get_new_light();
    *dst->data.light = *src->data.light;
    dst->type = src->type;
}

PRIVATE void
free_light(object *s)
{
    free((void *) s->data.light);
    s->type = NOSHAPE;
}

PRIVATE void
print_light(object *o)
{
#ifdef DEBUG
    struct light_data *l = o->data.light;

    switch (l->type) {
    case L_NORMAL:
	dprintf("pointsource\n");
	break;
	break;
    case L_FOGSOURCE:
	dprintf("fog\n");
	break;
    default:
	assert(FALSE);
    }
    print_c("color", l->lcolor);
    print_v("origin", l->org);
    print_v("direction", l->direction);
    dprintf("attenuation %f, radius  %f\n",
	    l->attenuation_exp, l->radius);
    dprintf("tightness %f ", l->tightness);

#endif				/* DEBUG */
}


PUBLIC void
print_light_list(void)
{
    int             i;

    for (i = 0; i < light_list_len; i++) {
	dprintf("light {\n");
	print_light(light_list[i]);
	dprintf("}\n");
    }
}

PRIVATE int     maxlightlist = 0;

/* add light to list of lights */
PRIVATE void
precompute_light(object *o)
{
    if (light_list_len >= maxlightlist) {
	maxlightlist = 2 * maxlightlist + 1;
	light_list = xrealloc(light_list, maxlightlist * sizeof(object *), "light list");
    }
    light_list[light_list_len++] = o;

    my_methods.howmuch++;
    global_stats.prims++;
}

PRIVATE struct methods my_methods =
{
    all_light_intersections,
    light_normal,
    copy_light,
    inside_light,
    rotate_light,
    translate_light,
    scale_light,
    free_light,
    print_light,
    precompute_light,
    "lightsource"
/*
 * hit and test are variables relating to the shadowray tests, check out
 * intersect.c
 */
};

PUBLIC object  *
get_new_light_object(void)
{
    object         *o;

    o = get_new_object();

    o->data.light = get_new_light();
    o->methods = &my_methods;
    o->type = LIGHTSOURCE;
    o->flags |= O_SMALL;

    return o;
}

PUBLIC color
light_color(struct light_data *l, vector a)
{
    vector          d;
    Flt             intensity,
                    dot;

    if (l->tightness) {
	vsub(d, a, l->org);
	norm(d, d);

	dot = vdot(d, l->direction);
	if (dot < 0)		/* this can happen, if the lighted point is not in the direction of the spotlight */
	    dot = 0;

	intensity = pow(dot, l->tightness);
	return scolorproduct(intensity, l->lcolor);
    } else
	return l->lcolor;
}
