/*
 * objects.c -- standard object manipulations
 *
 * (c) 1993, 1994 Han-Wen Nienhuys. <hanwen@stack.urc.tue.nl>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ray.h"
#include "proto.h"
#include "extern.h"

/*
 * intersect a ray with an object.
 */
PUBLIC bool
intersect_object(ibuf * globalq, object *o, struct ray *r, int flags, bool *isinside)
{
    struct ray      transray;
    bool            hit;
    object         *op;

    global_stats.objtest++;

    /* mailboxes */
    if ((flags & USEMAIL) && o->last_ray_id == r->rayid)
	return FALSE;
    else
	o->last_ray_id = r->rayid;

    /* the O_NO_XXX flags */
    if (r->type & o->flags)
	return FALSE;

    /* needed for correct refraction */
    if (o->flags & O_TRANSPARENT) {
	flags |= CHKINSIDE;

	/* will be needed for attenuation of shadowrays */
	if (r->type == R_SHADOW)
	    flags |= CHKALL;
    }
    /* use speed in the ray. */
    if (o->speed) {
	transray = *r;
	svproduct(transray.pos, -Time, *o->speed);
	vadd(transray.pos, r->pos, transray.pos);

	/* tricky: throw away pointer to original input ray */
	r = &transray;
    }
    /* check for bounding shapes, walk them all */
    for (op = o->bound; !(flags & NO_CHKBOUNDS) && op != NULL; op = op->next) {
	ibuf           *q;
	bool            inside_bound;

	q = get_new_ibuf();

	/* Check bounding shape: */
	global_stats.boundtest++;

	/* don't translate bounds. */
	if (!intersect_object(q, op, r, CHKINSIDE, &inside_bound)) {
	    /* no intersection at all */

	    free_ibuf(q);
	    if (flags & CHKINSIDE)
		*isinside = inside_object(o, r->pos);

	    return FALSE;
	}
	if (inside_bound) {
	    free_ibuf(q);
	    continue;
	}
	/* found one! */
	if (!(flags & CHKALL) && min_param(q) >= min_param(globalq)) {	/* close enough? */
	    free_ibuf(q);	/* no, exit */
	    if (flags & CHKINSIDE)
		*isinside = inside_object(o, r->pos);

	    return FALSE;
	}
	free_ibuf(q);

	global_stats.boundhit++;/* And another bound has been succesful */
    }


    flags &= ~NO_CHKBOUNDS;

    if (o->clip != NULL) {

	/* do clippingshape */
	ibuf           *q;
	dqueue         *qp;
	vector          x;


	q = get_new_ibuf();

	if (o->inv_trans) {
	    struct ray      localray = *r;

	    transform_ray(&localray, o);
	    o->methods->all_intersections_method(q, o, &localray, flags | CHKALL, isinside);
	} else {
	    o->methods->all_intersections_method(q, o, r, flags | CHKALL, isinside);
	}

	hit = FALSE;

	first_entry(q);

	for (qp = next_entry(q); qp != NULL && (flags & CHKALL || !hit); qp = next_entry(q)) {
	    bool            inside;

	    /* find intersection point */
	    svproduct(x, qp->t, r->dir);
	    vadd(x, r->pos, x);

	    global_stats.cliptest++;

	    /*
	     * filter out the intersections which are inside of the clipping shape.
	     */
	    inside = TRUE;
	    for (op = o->clip; op != NULL && inside; op = op->next) {
		if (!inside_object(op, x))
		    inside = FALSE;
	    }

	    if (inside) {
		add_to_ibuf(globalq, flags, qp);
		global_stats.cliphit++;
		hit = TRUE;
	    }
	}

	if (flags & CHKINSIDE)
	    *isinside = inside_object(o, r->pos);

	free_ibuf(q);
    } else {

	if (o->inv_trans) {
	    struct ray      localray;

	    localray = *r;
	    transform_ray(&localray, o);

	    /* intersect the ray with object's shape */
	    hit = o->methods->all_intersections_method(globalq, o, &localray, flags, isinside);
	} else {
	    /* intersect the ray with object's shape */
	    hit = o->methods->all_intersections_method(globalq, o, r, flags, isinside);
	}
    }
    if (hit)
	global_stats.objhit++;
    return hit;
}

/* apply speed to one object */
PUBLIC void
speed_object(object *o, vector s)
{
    if (o->type == COMPOSITE)
	speed_composite(o, s);
    else {
	if (!o->speed)
	    o->speed = get_new_vector();
	vadd(*o->speed, s, *o->speed);
    }
}

/* append toadd to list starting with o. Both must have been allocated. */
PUBLIC void
append_to_object_list(object *o, object *toadd)
{
    object         *last;

    assert(o != NULL);

    for (; o != NULL; o = o->next)
	last = o;
    last->next = toadd;
}

/* add clipping shape from boundingshapes. */
PUBLIC void
add_bounds_to_clip(object *o)
{
    object         *srcp,
                   *dstp;

    if (o->bound != NULL) {	/* this object has a bound. */
	o->clip = get_new_object();

	dstp = o->clip;

	for (srcp = o->bound; srcp != NULL; srcp = srcp->next) {
	    copy_object(dstp, srcp);
	    dstp->daddy = o;
	    if (srcp->next != NULL)
		dstp->next = get_new_object();

	    dstp = dstp->next;
	}
    }
}


/* apply speed to a whole bunch of objects */
PUBLIC void
speed_object_list(object *start, vector s)
{
    for (; start != NULL; start = start->next)
	speed_object(start, s);
}


/* standard routine */
PUBLIC void
copy_object(object *dst, object *src)
{
    assert(dst != NULL && src != NULL);

    /* copy info */
    dst->type = NOSHAPE;

    dst->flags = src->flags;
    dst->methods = src->methods;
    dst->next = NULL;
    dst->daddy = NULL;

    /* copy pointers */
    if (src->text != NULL) {	/* check wether sources have a texture */
	if (dst->text == NULL)	/* source has texture; does dest has space for one? */
	    dst->text = get_new_texture();	/* no, so alloc space */
	copy_texture(dst->text, src->text);	/* now copy stuff */
    }
    /* copy the shape */
    src->methods->copy_method(dst, src);

    if (src->inv_trans) {
	dst->inv_trans = get_new_matrix();
	dst->trans = get_new_matrix();
	copy_matrix(*dst->inv_trans, *src->inv_trans);
	copy_matrix(*dst->trans, *src->trans);
    }
    if (src->speed) {
	dst->speed = get_new_vector();

	copy_vector(dst->speed, src->speed);
    }
    /* do bounds and clips */
    {
	object         *srcp,
	               *dstp;

	if (src->bound != NULL) {	/* this object has a bound. */
	    dst->bound = get_new_object();

	    for (dstp = dst->bound, srcp = src->bound; srcp != NULL; srcp = srcp->next) {
		copy_object(dstp, srcp);
		if (srcp->next != NULL)
		    dstp->next = get_new_object();
		dstp = dstp->next;
	    }
	}
	if (src->clip != NULL) {/* this object has a bound. */
	    dst->clip = get_new_object();

	    for (dstp = dst->clip, srcp = src->clip; srcp != NULL; srcp = srcp->next) {
		copy_object(dstp, srcp);
		if (srcp->next != NULL)
		    dstp->next = get_new_object();
		dstp = dstp->next;
	    }
	}
    }

}


/* precompute object */
PUBLIC void
precompute_object(object *o)
{
    object         *op;


    /* precompute parafernalia */
    precompute_object_list(o->bound, NULL);
    precompute_object_list(o->clip, NULL);

    /* allocate a bounding box, if necessary. */
    if (!o->bbox) {
	o->bbox = xmalloc(2 * sizeof(vector), "bbox");

	/* set bbox[0] and bbox[1] to infinity */
	setvector(o->bbox[0], -INFTY, -INFTY, -INFTY);
	vneg(o->bbox[1], o->bbox[0]);
    }
    /* ... Do It .. */
    o->methods->precompute_method(o);

    /* can I see clear now ? (check transparency) */
    for (op = o; op->text == NULL; op = op->daddy);

    if (color_length(op->text->refr_color) > EPSILON) {
	o->flags |= O_TRANSPARENT;
	if (!(o->flags & O_CLOSED))
	    warning("Transparent objects should be closed");
    }
    /* flip inside and outside */
    if ((o->flags & O_INVERTED)) {
	int             fi,
	                fo;

	fo = o->flags & O_FINITE_OUTSIDE;
	fi = o->flags & O_FINITE_INSIDE;

	if (fi)
	    o->flags |= O_FINITE_OUTSIDE;
	else
	    o->flags &= ~O_FINITE_OUTSIDE;

	if (fo)
	    o->flags |= O_FINITE_INSIDE;
	else
	    o->flags &= ~O_FINITE_INSIDE;
    }
    /* stats */
    if (o->bound != NULL)
	global_stats.bounds++;

    if (o->clip != NULL)
	global_stats.clips++;



    /* take intersection of Bboxes */
    for (op = o->bound; op != NULL; op = op->next) {
	if (!(op->flags & O_INFINITE))
	    o->flags &= ~O_INFINITE;
	else
	    warning("Bounding shape is infinite");
	bbox_intersection(&o->bbox[0], &o->bbox[1], &op->bbox[0], &op->bbox[1]);
    }

    /* take intersection of clips */
    for (op = o->clip; op != NULL; op = op->next) {
	if (!(op->flags & O_INFINITE))
	    o->flags &= ~O_INFINITE;
	bbox_intersection(&o->bbox[0], &o->bbox[1], &op->bbox[0], &op->bbox[1]);
    }

    /* incorporate speed into the bounds */
    if (o->speed) {
	vector          v1,
	                v2,
	                m1,
	                m2;

	svproduct(v1, time1, *o->speed);
	svproduct(v2, time2, *o->speed);

	vadd(m1, o->bbox[1], v1);
	vadd(m2, o->bbox[1], v2);
	o->bbox[1].x = MAX(m1.x, m2.x);
	o->bbox[1].y = MAX(m1.y, m2.y);
	o->bbox[1].z = MAX(m1.z, m2.z);

	vadd(m1, o->bbox[0], v1);
	vadd(m2, o->bbox[0], v2);

	o->bbox[0].x = MIN(m1.x, m2.x);
	o->bbox[0].y = MIN(m1.y, m2.y);
	o->bbox[0].z = MIN(m1.z, m2.z);

    } {
	vector          fudge;

	vsub(fudge, o->bbox[1], o->bbox[0]);
	svproduct(fudge, BOUNDFUDGE, fudge);
	/* enlarge bbox slightly */
	vadd(o->bbox[1], o->bbox[1], fudge);
	vsub(o->bbox[0], o->bbox[0], fudge);
    }


    /* After generating the bbox, the trans matrix isn't needed. (the inv_trans is enough) */
    free_matrix(o->trans);
    global_stats.objects++;
}

/* precompute a linked list of objects */
PUBLIC void
precompute_object_list(object *o, object *daddy)
{
    for (; o != NULL; o = o->next) {
	precompute_object(o);
	if (daddy) {
	    assert(daddy == o->daddy);
	}
    }
}

/* initialize the object passed by p */
PUBLIC void
init_object(object *p)
{
    p->speed = NULL;
    p->bbox = NULL;
    p->text = NULL;
    p->type = NOSHAPE;

    p->daddy = p->next = p->bound = p->clip = NULL;
    p->trans = p->inv_trans = NULL;

    p->methods = NULL;
    p->flags = 0;
    p->last_ray_id = -1;
}

/* alloc and init an object */
PUBLIC object  *
get_new_object(void)
{
    object         *p;

    p = ALLOC(object);


    init_object(p);
    return p;
}

/* get normal to object. Handles inverse transforms */
PUBLIC vector
object_normal(struct intersect *i, vector loc)
{
    object         *o = i->q_ent.obj;
    vector          normal;

    if (o->inv_trans)
	loc = mvproduct(*o->inv_trans, loc);

    normal = o->methods->normal_method(i, loc);
    if (o->inv_trans)
	normal = transform_normal(*o->inv_trans, normal);

    norm(normal, normal);

    return normal;
}

/* is "loc" inside "o"? Handles inverse transforms */
PUBLIC bool
inside_object(object *o, vector loc)
{
    bool            inside;

    if (o->inv_trans)
	loc = mvproduct(*o->inv_trans, loc);

    inside = o->methods->inside_method(o, loc);

    if (o->flags & O_INVERTED)
	return !inside;
    else
	return inside;
}

/* std routine: rotate an object */
PUBLIC void
rotate_object(object *o, matrix rotmat)
{
    if (o->text != NULL)
	rotate_texture(o->text, rotmat);

    o->methods->rotate_method(o, rotmat);

    if (o->speed)
	rotate_vector(o->speed, rotmat);

    if (o->bound != NULL)
	rotate_object_list(o->bound, rotmat);
    if (o->clip != NULL)
	rotate_object_list(o->clip, rotmat);
}

/* translate an object */
PUBLIC void
translate_object(object *op, vector t)
{
    if (op->text != NULL)
	translate_texture(op->text, t);

    op->methods->translate_method(op, t);

    if (op->bound != NULL)
	translate_object_list(op->bound, t);
    if (op->clip != NULL)
	translate_object_list(op->clip, t);
}

/* and scale it */
PUBLIC void
scale_object(object *o, vector s)
{
    o->methods->scale_method(o, s);

    if (o->text != NULL)
	scale_texture(o->text, s);

    if (o->speed)
	vcproduct(*o->speed, *o->speed, s);

    if (o->bound != NULL)
	scale_object_list(o->bound, s);
    if (o->clip != NULL)
	scale_object_list(o->clip, s);
}


/* scale the object->data.XXXX with transform matrices */
PUBLIC void
generic_scale_object(object *o, vector s)
{
    matrix          m;

    if (o->inv_trans == NULL) {
	o->inv_trans = get_new_matrix();
	o->trans = get_new_matrix();
    }
    inv_scale_matrix(m, s);
    mmproduct(*o->inv_trans, *o->inv_trans, m);
    scale_matrix(m, s);
    mmproduct(*o->trans, m, *o->trans);
}

/* translate the object->data.XXXX with transform matrices */
PUBLIC void
generic_translate_object(object *o, vector t)
{
    matrix          m;

    if (o->inv_trans == NULL) {
	o->inv_trans = get_new_matrix();
	o->trans = get_new_matrix();
    }
    inv_translate_matrix(m, t);
    mmproduct(*o->inv_trans, *o->inv_trans, m);
    translate_matrix(m, t);
    mmproduct(*o->trans, m, *o->trans);
}

/* rotate the object->data.XXXX with transform matrices */
PUBLIC void
generic_rotate_object(object *o, matrix m)
{
    matrix          r;

    if (o->inv_trans == NULL) {
	o->inv_trans = get_new_matrix();
	o->trans = get_new_matrix();
    }
    mmproduct(*o->trans, m, *o->trans);
    transpose_matrix(r, m);
    mmproduct(*o->inv_trans, *o->inv_trans, r);
}

/* give back memory used by an object */
PUBLIC void
free_object(object *op)
{
    op->methods->free_method(op);
    if (op->text != NULL)
	free_texture(op->text);
    if (op->bound != NULL)
	free_object_list(op->bound);
    if (op->clip != NULL)
	free_object_list(op->clip);
    if (op->inv_trans != NULL)
	free(op->inv_trans);
    if (op->trans != NULL)
	free(op->trans);
    if (op->bbox != NULL)
	free(op->bbox);

    free((void *) op);
}

/* zap a list of objects */
PUBLIC void
free_object_list(object *p)
{
    object         *next;

    for (; p != NULL; p = next) {
	next = p->next;
	free_object(p);
    }
}

/* rotation */
PUBLIC void
rotate_object_list(object *p, matrix rotmat)
{
    for (; p != NULL; p = p->next)
	rotate_object(p, rotmat);
}

/* scaling */
PUBLIC void
scale_object_list(object *p, vector r)
{
    for (; p != NULL; p = p->next)
	scale_object(p, r);
}

/* translation */
PUBLIC void
translate_object_list(object *p, vector r)
{
    for (; p != NULL; p = p->next)
	translate_object(p, r);
}

/* print out object flags */
PRIVATE void
print_flags(int flags)
{
#ifdef DEBUG
    if (!flags)
	return;
    dprintf("object flags: ");
    if (flags & O_NO_SHADOW)
	dprintf("noshad ");
    if (flags & O_NO_REFLECT)
	dprintf("norefl");
    if (flags & O_NO_REFRACT)
	dprintf("norefr ");
    if (flags & O_NO_EYE)
	dprintf("noeye ");
    if (flags & O_TRANSPARENT)
	dprintf("transp ");
    if (flags & O_INFINITE)
	dprintf("infin ");
    if (flags & O_SMALL)
	dprintf("invis ");
    if (flags & O_CLOSED)
	dprintf("closed ");
    if (flags & O_FINITE_INSIDE)
	dprintf("FinIns ");
    if (flags & O_FINITE_OUTSIDE)
	dprintf("FinOut ");
    if (flags & O_INVERTED)
	dprintf("Inv ");
    dprintf("\n");
#endif				/*DEBUG*/
}

PUBLIC void
print_object(object *o)
{
#ifdef DEBUG
    dprintf("object {");
    if (o->bound != NULL) {
	dprintf("BOUNDINGSHAPES:");
	print_object_list(o->bound);
    }
    if (o->clip != NULL) {
	dprintf(" CLIPPINGSHAPES:");
	print_object_list(o->clip);
    }
    print_flags(o->flags);
    if (!(o->flags & O_SMALL) && !(o->flags & O_INFINITE) && o->bbox) {
	dprintf("BBOX:\n");
	print_v("bound_min", o->bbox[0]);
	print_v("bound_max", o->bbox[1]);
    }
    if (o->inv_trans != NULL) {
	dprintf("Inverse Transformation\n");
	print_matrix(*o->inv_trans);
    }
    dprintf("shape type %d, %s {\n", o->type, o->methods->name);

    o->methods->print_method(o);
    dprintf("}\n");
    if (o->text != NULL)
	print_texture(o->text);
    if (o->speed)
	print_v("speed", *o->speed);
    dprintf("}\n\n");
#endif				/* DEBUG */
}

PUBLIC void
short_print_object(object *o)
{
#ifdef DEBUG
    assert(o != NULL);
    if (o->bbox) {
	dprintf("BBOX:\n");
	print_v("bound_min", o->bbox[0]);
	print_v("bound_max", o->bbox[1]);
    }
    dprintf("shape type %d, %s\n", o->type, o->methods->name);
#endif				/* DEBUG */
}

PUBLIC void
print_object_list(object *o)
{
#ifdef DEBUG
    for (; o != NULL; o = o->next)
	print_object(o);
#endif				/* DEBUG */
}
