/*
 * objects.h --type definitions for primitives and nonprimitives.
 */

/* quadric equation coefficients */
struct quadric_data {
    Flt             xx,
                    yy,
                    zz,
                    xy,
                    yz,
                    xz,
                    x0,
                    y0,
                    z0,
                    a;
};

/* the sphere structure */
struct sphere_data {
    vector          center;
    Flt             radius,
                    rsq;	/* radius squared */
};

struct extrusion_data {
    object         *shape;	/* the shape to be extruded */
    Flt             height;	/* height of the extrusion */
    bool            locap,
                    hicap;	/* are bottom/top to be capped ? */
};

struct conic_data {
    object         *shape;	/* the shape to be extruded */
    Flt             height;	/* height of the conic extrusion */
    vector          apex;	/* apex of cone */
    bool            locap,
                    hicap;	/* are bottom/top to be capped ? */
};

struct cone_data {
    vector          p1,
                    p2;
    Flt             r1,
                    r2,
                    slope,
                    mov1,
                    mov2;
    vector          axis;
    bool            locap,
                    hicap;	/* are bottom/top to be capped ? */
};

/* convexM-^H polygon */
struct polygon_data {
    int             no_vertices;/* number of vertices */
    vector         *vertices;	/* vertices themselves */
    vector         *clips;	/* inward facing normals */
    vector          normal;	/* normal of the polygon */
    Flt             mov;	/* mov (see plane_data) */
    bool            convex;
    char            project_to;
};

/* a (optionally smooth) triangle */
struct triangle_data {
    vector          n;		/* normal of plane defined by vertices */
    Flt             mov;	/* translate of plane */
    char            project_to;
    vector          vertices[3];/* vertices of the triangle */
    vector         *normals;	/* normals; if allocated, then they exist. */
    vector          linfunc1,
                    linfunc2;	/* linear functionals, which compute the components of a vector in the basis formed by
				 * vert[0]-vert[2], vert[1] - vert[2] */
};

/* plane info */
struct plane_data {
    vector          n;		/* the normal */
    Flt             mov;	/* translation parallel to normal */
};

struct disc_data {
    vector          n;		/* the normal */
    Flt             mov;	/* translation parallel to normal */
    vector          center;
    Flt             R,		/* radii */
                    r,
                    Rsq,	/* radii squared */
                    rsq;
};

/* a PoV box */
struct box_data {
    vector          p1,
                    p2;		/* point with minimum/maximum coordinates */
};

/* lighttypes */
enum {
    L_NORMAL,
    L_FOGSOURCE,
};

/* light source info */
struct light_data {
    vector          org,	/* the place at which the light is. */
                    direction;

    color           lcolor;	/* color of the light */

    int             type;

    Flt             attenuation_exp,	/* exponent E for intensity = 1/ldist^E */
                    tightness,	/* for spotlights */
                    radius;	/* radius, used for falloff and penumbra */
};

struct torus_data {		/* for a torus. */
    Flt
    Rad,
    rad;			/* major/minor radius */
    bool            use_sturm;	/* sturm seqs? */
};

struct superq_data {
    vector          powvect;	/* power of norm ||.|| */
};

struct algebraic_data {
    struct poly_instruction *code;	/* code needed to generate the one-var polynomial */
    int             codesize;	/* length of code */
    bool            usesturm;	/* calculate using sturm seqs? */
};


/* types of optimisation */
enum {
    NONE,			/* use parent's optimisation */
    MANUAL,			/* user specified BVH */
    BSP,			/* build BSP of contents */
};

/* list of objects to be optimised */
struct optilist {
    object        **list;
    int             len,
                    max;
    vector          bmin,
                    bmax;
};

struct composite_data {
    object         *contents;	/* linked list of the elements */
    void           *optidata;
    bool            (*all_intersections_method) (ibuf * q, object *op, struct ray * r, int flags, bool *b);
    void            (*print_optimisation) (object *o);
    int             optitype;
};


/* for the shape->type variable */
enum {
    NOSHAPE, SPHERE, QUADRIC, PLANE, LIGHTSOURCE, BOX,
    TORUS, CSGUNION, CSGINTER, COMPOSITE, ALGEBRAIC, SUPERQ,
    POLYGON, EXTRUSION, TRIANGLE, DISC, CONIC, CONE
};

/* entry for stack of polynomial operations */
struct poly_instruction {
    int             type;	/* what kind of instruction */
    union {
	Flt             factor;	/* numeric factor */
	char            op;	/* operator */
    }
    data;
};

/* polynomial operations */
enum {
    POLY_CONS,			/* push constant */
    POLY_OP,			/* do operation on stack */
    POLY_X,			/* push X */
    POLY_Y,			/* push Y */
    POLY_Z,			/* push Z */
};
union shape_data {		/* the data. */
    struct plane_data *plane;
    struct sphere_data *sphere;
    struct quadric_data *quadric;
    struct light_data *light;
    struct box_data *box;
    struct torus_data *torus;
    struct algebraic_data *algebraic;
    struct superq_data *superq;
    struct polygon_data *polygon;
    struct extrusion_data *extrusion;
    struct triangle_data *triangle;
    struct disc_data *disc;
    struct conic_data *conic;
    object         *CSG;
    struct composite_data *composite;
};
