/*
 * proto.h -- function prototypes
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifndef DEBUG			/* make assert() calls stop */
#define NDEBUG
#endif				/* DEBUG */
#include <assert.h>

#define DEBUGTOK	0x01	/* print tokens as they are read */
#define DEBUGPARS	0x02	/* turn yydebug on */
#define DEBUGRUNTIME	0x04	/* print memory statistics if something is alloced. */
#define DEBUGINT	0x08	/* print everything in scene before tracing and after parsing */
#define DEBUGDECL	0x16	/* print everything which is declared */
#define DEBUGSELFCHECK  0x32	/* miscellaneous self checks. */

#ifndef PATHSEP
#define PATHSEP "/"
#endif

/* miscellaneous routines */
PUBLIC void     transform_ray(struct ray *r, object *o);
PUBLIC void     close_input(void);
PUBLIC void     raytrace(char *fname);
PUBLIC void     readfile(char *fname);
PUBLIC void     set_input(char *fname);
PUBLIC void     abort_trace(int);
PUBLIC FILE    *open_file(char *name, char *mode);
PUBLIC void     close_file(FILE * f);
PUBLIC int      getopt(int nargc, char *const *nargv, char const *ostr);

/* device dep. code */
PUBLIC void     check_abort(void);
PUBLIC void     init_machine(void);
PUBLIC void     initialize(void);
PUBLIC void     timer_start(void);
PUBLIC long     timer_stop(void);

/* errors */
PUBLIC void     warning(char *format,...);
PUBLIC void     errormsg(char *format,...);
PUBLIC void    *xmalloc(size_t s, char *type);
PUBLIC void    *xrealloc(void *ptr, size_t s, char *type);
PUBLIC char    *xstrdup(char *s);

/* stats */
PUBLIC void     start_stats(void);
PUBLIC void     set_defaults(void);
PUBLIC void     print_all_shape_stats(char *format, object *p);
PUBLIC void     print_image_map(struct image_map *p);

/* debug */
PUBLIC void     print_v(char *s, vector v);
PUBLIC void     print_matrix(matrix c);
PUBLIC void     print_poly(spoly *p);
PUBLIC void     print_queue(dqueue * root);
PUBLIC void     print_c(char *s, color c);
PUBLIC void     dprintf(char *format,...);
PUBLIC void     print_globals(void);

PUBLIC void     clean_optimisation_list(struct optilist *l);
PUBLIC void     add_to_optilist(struct optilist *thelist, object *op);
PUBLIC void     build_list(object *o, struct optilist *thelist);
PUBLIC void     print_optimisation_list(struct optilist *thelist);
PUBLIC struct optilist *get_new_optimisation_list(void);
PUBLIC void     free_optimisation_list(struct optilist *l);

/* colors. */
PUBLIC void     scale_color(color *c, Flt fact);
PUBLIC color    color_add(color a, color b);
PUBLIC color    color_mul(color a, color b);
PUBLIC color    scolorproduct(Flt fact, color c);
PUBLIC color    color_scaled(color *c, Flt fact);
PUBLIC void     copy_color(color *dst, color *src);
PUBLIC color    normcol(color col);
PUBLIC Flt      color_length(color c);
PUBLIC color   *get_new_color(void);
PUBLIC void     free_color(color *c);

/*
  function prototypes for all global functions
 */

/* include macros */
#include "macro.h"
#include "vector.h"
#include "ibuf.h"

/* math routines */
PUBLIC int      sturm_solve_rt_poly(spoly *thepoly, Flt * roots, bool chkall);
PUBLIC int      solve_rt_poly(spoly *thepoly, Flt * roots, bool chkall);
PUBLIC Flt      poly_eval(Flt x, int ord, Flt * coef);
PUBLIC Flt      degtorad(Flt deg);
PUBLIC Flt      radtodeg(Flt rad);
PUBLIC Flt      frand(void);
PUBLIC Flt      rand1(void);
PUBLIC Flt      Gauss_rand(void);
PUBLIC void     torotv(vector *v);
PUBLIC bool     isvnull(vector a);
PUBLIC void     copy_vector(vector *dst, vector *src);
PUBLIC void     update_min_and_max(vector *min, vector *max, vector point);
PUBLIC void     bbox_intersection(vector *bigmin, vector *bigmax, vector *min, vector *max);
PUBLIC void     bbox_union(vector *bigmin, vector *bigmax, vector *min, vector *max);
PUBLIC void     do_box_bound(vector *bmin, vector *bmax, matrix trans, vector p1, vector p2);
PUBLIC Flt      safe_arctangent(Flt a, Flt b);

/* matrices */
PUBLIC vector   transform_normal(matrix A, vector n);
PUBLIC void     split_inv_trans(matrix A, vector *trans, vector *scal, matrix rot);
PUBLIC void     copy_matrix(matrix dst, matrix src);
PUBLIC void     rotate_matrix(matrix m, vector v);
PUBLIC void     mmproduct(matrix product, matrix left, matrix right);
PUBLIC void     transpose_matrix(matrix tr, matrix org);
PUBLIC vector   mvproduct(matrix A, vector x);
PUBLIC vector   transform_direction(matrix A, vector x);
PUBLIC void     invert_trans(matrix out, matrix in);
PUBLIC void     make_rotation_matrix(matrix m, vector v);

PUBLIC void     null_matrix(matrix m);
PUBLIC void     unit_matrix(matrix m);
PUBLIC void     inv_rotate_matrix(matrix m, vector r);
PUBLIC void     scale_matrix(matrix m, vector s);
PUBLIC void     inv_scale_matrix(matrix m, vector s);
PUBLIC void     translate_matrix(matrix m, vector t);
PUBLIC void     inv_translate_matrix(matrix m, vector t);
PUBLIC void     free_matrix(matrix *f);


/* the real tracing stuff */
PUBLIC Flt      find_roughness(Flt beta, int type);
PUBLIC color    shade(struct intersect *i, grid_tree * g, Flt imp);
PUBLIC color    bgcolor(struct ray *r);
PUBLIC color    fogsources(struct ray *r, Flt endpoint);
PUBLIC color    trace(struct ray r, grid_tree * g, Flt imp);
PUBLIC struct intersect intersect(struct ray r);

PUBLIC void     sample_ray(struct ray *r, Flt * filter_factor, Flt theta, grid_tree * g);
PUBLIC bool     shadow_test(ibuf * queue, struct ray r, Flt ldistance, int index, grid_tree * g);

PUBLIC ibuf    *get_new_ibuf(void);
PUBLIC void     free_ibuf(ibuf * i);
PUBLIC void     add_to_ibuf(ibuf * i, int flags, dqueue * q);
PUBLIC void     sort_ibuf(ibuf * i);
PUBLIC void     print_ibuf(ibuf * i);


/*
 * intersection routines; global so they can be used for special bounding
 * shapes
 */
PUBLIC int      intersect_rawbox(vector p1, vector p2, struct ray *r, Flt * hits);
PUBLIC int      intersect_sphere(vector center, Flt rsq, struct ray *r, Flt * hits);
PUBLIC bool     intersect_object(ibuf * globalq, object *o, struct ray *r, int flags, bool *isinside);

/* precompute_xxx. They compute constants in the xxx_data structure */
PUBLIC void     precompute_object_list(object *o, object *daddy);
PUBLIC void     precompute_object(object *o);
PUBLIC void     precompute_globals(void);

/* polygons */
PUBLIC void     add_vertex_to_polygon(object *o, vector vert);

/* the get_new_xxx routines. They alloc xxx, and initialize it */
PUBLIC Flt     *get_new_float(void);
PUBLIC matrix  *get_new_matrix(void);

/* get_new_xxx_object */
PUBLIC object  *get_new_disc_object(void);
PUBLIC object  *get_new_smooth_triangle_object(void);
PUBLIC object  *get_new_triangle_object(void);
PUBLIC object  *get_new_extrusion_object(void);
PUBLIC object  *make_cylinder_object(vector p1, vector p2, Flt rad);
PUBLIC object  *get_new_torus_object(void);
PUBLIC object  *get_new_object(void);
PUBLIC object  *get_new_superq_object(void);
PUBLIC object  *get_new_sphere_object(void);
PUBLIC object  *get_new_quadric_object(void);
PUBLIC object  *get_new_plane_object(void);
PUBLIC object  *get_new_light_object(void);
PUBLIC object  *get_new_box_object(void);
PUBLIC object  *get_new_CSGinter_object(void);
PUBLIC object  *get_new_CSGunion_object(void);
PUBLIC object  *get_new_algebraic_object(void);
PUBLIC object  *get_new_polygon_object(void);
PUBLIC object  *get_new_conic_object(void);

/* free_xxx routines. They free xxx, and xxx's sub-data */
PUBLIC void     free_float(Flt * f);

/* vectors */
PUBLIC void     rotate_vector(vector *v, matrix rotmat);
PUBLIC void     free_vector(vector *v);
PUBLIC vector  *get_new_vector(void);

/* polynomials */
PUBLIC void     poly_neg(spoly *dst, spoly *src);
PUBLIC void     poly_multiply(spoly *dest, spoly *p1, spoly *p2);
PUBLIC void     poly_power(spoly *dst, int exponent, spoly *src);
PUBLIC void     poly_copy(spoly *dst, spoly *src);
PUBLIC void     poly_clean(spoly *p);
PUBLIC void     poly_add(spoly *dst, spoly *p1, spoly *p2);
PUBLIC void     poly_scalmult(spoly *d, Flt fact);
PUBLIC void     poly_sub(spoly *dst, spoly *p1, spoly *p2);
PUBLIC void     poly_real_clean(spoly *p);

/* cameras */
PUBLIC Flt      to_fov_coeff(struct camera *c, Flt angle);
PUBLIC void     point_camera(struct camera *c, vector lookat);
PUBLIC struct camera *get_new_camera(void);
PUBLIC void     free_camera(struct camera *c);
PUBLIC void     translate_camera(struct camera *c, vector t);
PUBLIC void     rotate_camera(struct camera *c, matrix rotmat);
PUBLIC void     scale_camera(struct camera *c, vector s);
PUBLIC void     copy_camera(struct camera *dst, struct camera *src);
PUBLIC void     print_camera(struct camera *c);

/* objects */
PUBLIC void     print_object_list(object *o);
PUBLIC void     print_object(object *obj);
PUBLIC void     short_print_object(object *obj);
PUBLIC void     free_object(object *op);
PUBLIC void     free_object_list(object *p);
PUBLIC void     rotate_object(object *op, matrix rotmat);
PUBLIC void     translate_object(object *op, vector t);
PUBLIC void     scale_object(object *op, vector s);
PUBLIC void     rotate_object_list(object *p, matrix rotmat);
PUBLIC void     scale_object_list(object *p, vector r);
PUBLIC void     translate_object_list(object *p, vector r);
PUBLIC void     speed_object_list(object *start, vector s);
PUBLIC void     copy_object(object *dst, object *src);
PUBLIC vector   object_normal(struct intersect *i, vector loc);
PUBLIC bool     inside_object(object *o, vector loc);

PUBLIC void     generic_scale_object(object *o, vector s);
PUBLIC void     generic_translate_object(object *o, vector t);
PUBLIC void     generic_rotate_object(object *o, matrix m);

PUBLIC void     append_to_object_list(object *o, object *toadd);
PUBLIC void     add_bounds_to_clip(object *o);


/* textures! */
PUBLIC void     fill_texture(struct texture_data *tp, struct texture_data *orgtext, struct intersect *i);
PUBLIC struct texture_data *get_new_texture(void);
PUBLIC void     free_texture(struct texture_data *t);
PUBLIC struct texture_data *copy_texture(struct texture_data *dst, struct texture_data *src);
PUBLIC void     translate_texture(struct texture_data *t, vector v);
PUBLIC void     scale_texture(struct texture_data *t, vector s);
PUBLIC void     rotate_texture(struct texture_data *t, matrix rotmat);
PUBLIC void     print_texture(struct texture_data *tx);

/* imagemaps */
PUBLIC struct pixel_data *get_TGA_pixels(char *fn);
PUBLIC struct pixel_data *get_GIF_pixels(char *fn);
PUBLIC color    get_image_map_color(struct image_map *p, struct intersect *i);
PUBLIC struct image_map *get_new_image_map(void);
PUBLIC void     copy_image_map(struct image_map *dst, struct image_map *src);
PUBLIC void     inverse_spheremap(Flt * u, Flt * v, vector x);
PUBLIC void     inverse_planemap(Flt * u, Flt * v, vector x);
PUBLIC void     inverse_torusmap(Flt * u, Flt * v, vector x);
PUBLIC void     inverse_cylindermap(Flt * u, Flt * v, vector x);
PUBLIC void     load_gif(struct pixel_data *image, FILE * f);
PUBLIC void     free_image_map(struct image_map *p);

/* composites */
PUBLIC object  *get_new_composite_object(void);
PUBLIC void     speed_composite(object *o, vector spd);
PUBLIC void     speed_object(object *o, vector s);
PUBLIC void     add_to_composite(object *c, object *o);
PUBLIC void     add_to_CSG(object *c, object *o);

PUBLIC void     init_BSP(object *o);


/* lights */
PUBLIC color    light_color(struct light_data *l, vector a);
PUBLIC void     print_light_list(void);

PUBLIC void     add_grid(grid_tree * g);
PUBLIC void     sample_grid(Flt * u, Flt * v, Flt * f, grid_tree * g, int raytype);
PUBLIC void     init_global_grid(void);
PUBLIC void     print_grid(void);
