/*
 * ray.h --  types, constants.
 *
 * (c) 1993, 1994 Han-Wen Nienhuys, <hanwen@stack.urc.tue.nl>
 *
 * Based on work by George Kyriazis,  1988
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef	FALSE
#define	FALSE	0
#endif

#ifndef	TRUE
#define	TRUE	1
#endif

#define EXTERN extern		/* so space can easily be allocated, see initalize.c */

#define PRIVATE static		/* local to a module */
#define PUBLIC			/* global function */

#define	EPSILON	1e-10		/* ridiculously small */
#define INFTY	1e20		/* ridiculously big, unity of MIN()  */
#define BOUNDFUDGE 1e-3		/* make bounds this amount too big */
#define DEFAULT_TOLERANCE 1e-3

#define	ALG_TOLERANCE DEFAULT_TOLERANCE / 2	/* max error for superqs, and algebraics */

/* constants and bounds for the program */
#define	DEFAULTMAXLEVEL	3	/* default maximum recursion level */
#define	MAX_ORDER	12	/* maximum polynomial order */
#define MAXMAPS		20	/* maximum number of maps to keep track of */
#define BUFFERSIZE	10240	/* size of input file buffers */

typedef double  LFlt;
typedef double  Flt;
typedef signed char bool;	/* a boolean variable */
typedef Flt matrix[4][4];	/* a 4x4 matrix, 3x3 for rotation, an extra column for translation */

typedef unsigned char byte;

/* guess what */
typedef struct {
    Flt             x,
                    y,
                    z;
}

vector;

/* the color structure */
typedef struct {
    Flt             r,
                    g,
                    b;
}

color;


/**************************************************************************/
/* the intersection structure */

/* the ray structure */
struct ray {
    vector          pos;	/* origin */
    vector          dir;	/* direction */
    char            type;	/* what kind, see above */

    long            rayid;	/* number of ray */
    Flt             maxt,	/* biggest value to look for (if it is more, than don't check intersection details */
                    mint;	/* smallest t values to look for. (if it is less, then exit.. Useful for shadow rays */
};

typedef struct _obj object;

/* Depth queue */
typedef struct _dqstruct dqueue;

struct _dqstruct {
    object         *obj,	/* the object which was hit. */
                   *obj2;	/* needed for extrusion */
    Flt             t;		/* where in the ray */
    bool            entering;	/* entering or exiting? */
};


/* structure for passing intersection-depth info */
struct intersect_buf {
    dqueue         *list,	/* array of intersections */
                    minq;	/* closest intersection */
    int             len,	/* length of list */
                    max,	/* amount of mem alloced for list */
                    idx;	/* temporary variable */
    bool            sorted,	/* sorted? */
                    free;	/* vacant buffer ? */
};

typedef struct intersect_buf ibuf;

/* structure for passing intersection-point info */
struct intersect {
    dqueue          q_ent;	/* the minimum intersection */
    struct ray      r;		/* the ray */
    vector          n,		/* the normal at that point */
                    x,		/* the intersection point in world coordinates (with speed) */
                    objectx,	/* intersection in shape coordinates */
                    objectn;	/* normal in lcs */
};

/* ray types */
enum {
    R_SHADOW = 0x01,
    R_REFLECT = 0x02,
    R_REFRACT = 0x04,
    R_EYE = 0x08,
};

enum {
    CHKALL = 0x01,		/* generate all intersections */
    CHKINSIDE = 0x02,		/* determine wether */
    USEMAIL = 0x04,		/* update mailboxes? */
    NO_CHKBOUNDS = 0x08,	/* check the bounding shapes ? */
    NO_TRANSUPDATE = 0x10	/* don't store transparent objects as minimal intersections */
};

/* object flags types */
enum {
    O_NO_SHADOW = R_SHADOW,	/* invisible for shadowrays */
    O_NO_REFLECT = R_REFLECT,	/* invisible for reflectionrays */
    O_NO_REFRACT = R_REFRACT,	/* invisible for refraction */
    O_NO_EYE = R_EYE,		/* invisible for eyerays */
    O_TRANSPARENT = 0x10,	/* with tranparent texture (needs all intersections/inside check) */
    O_INFINITE = 0x20,		/* objects boundary is infinite (eg plane) */
    O_SMALL = 0x40,		/* invisible (lightsource) */
    O_CLOSED = 0x80,		/* closed surface */
    O_FINITE_INSIDE = 0x100,	/* object's inside is infinite, eg sphere, NOT inverted sphere */
    O_FINITE_OUTSIDE = 0x200,	/* object's inside is infinite, eg inverted sphere */
    O_INVERTED = 0x400,		/* for CSG purposes. */
};

typedef struct _grid_tree grid_tree;

enum {
    NO_FILTER,
    GAUSS_FILTER
};

struct _grid_tree {
    grid_tree      *reflect_grid,	/* children */
                   *refract_grid;
    int            *indices,	/* indices to global grid */
                   *shadowindices,	/* indices to shadow grid */
                    last,
                    shadowlast,
                    recursion;
    object        **scache;	/* shadow cache */
};


/* standard manipulators for objects */
struct methods {
    bool            (*all_intersections_method) (ibuf * q, object *op, struct ray * r, int flags, bool *b);
    vector          (*normal_method) (struct intersect *, vector);
    void            (*copy_method) (object *, object *);
    bool            (*inside_method) (object *, vector);
    void            (*rotate_method) (object *, matrix rotmat);
    void            (*translate_method) (object *, vector);
    void            (*scale_method) (object *, vector);
    void            (*free_method) (object *);
    void            (*print_method) (object *);
    void            (*precompute_method) (object *);
    char           *name;
    long            test,
                    hit,
                    howmuch;
};

/* some statistics variables */
struct global_stat_struc {
    long
                    raytest,	/* total number of ray/intersection tests */
                    rayintersections,	/* number of rays hitting something */
                    objtest,	/* number of objects tested. */
                    objhit,	/* total no. objects hit */
                    shadowtest,	/* total number shadow ray tests */
                    reflected,	/* total number of reflected rays */
                    refracted,	/* total number of refracted rays */
                    eyerays,
                    boundtest,
                    boundhit,
                    cliptest,
                    cliphit,
                    cachetest,
                    cachehit,
                    cachemiss,
                    objects,
                    bounds,
                    clips,
                    prims,
                    nonprims;
};

/* statistics for a single line. */
struct line_stat_struc {
    long
                    raytest,	/* rays */
                    rayintersections,	/* rays hitting something */
                    shadowtest,	/* shadow test */
                    reflected,	/* reflected rays */
                    refracted,	/* refracted rays */
                    eyerays;	/* eye rays */
    short           maxdepth;	/* maximum depth reached. */
};

enum {				/* texturing. */
    T_EGAL,			/* smooth */
    T_BLEND,			/* texture blend (not yet) */
    T_IMAGEMAP,			/* image mapping */
    T_COLORMAP,			/* color mapped texture: perlin style (not yet) */
};

enum {
    PHONG_DISTRIBUTION,
    BLINN_DISTRIBUTION,
    GAUSS_DISTRIBUTION,
    COOK_DISTRIBUTION,
    REITZ_DISTRIBUTION,
};

#include "objects.h"

/* structure to hold texture info */
struct texture_data {
    int             type,
                    shadingtype,
                    distribution;

    union {
	struct image_map *image;
    } exttext;			/* extended info */
    color          *PoVcolor;	/* for compatibility.. */

    color           refl_color,	/* reflective color */
                    refr_color,	/* refractive color */
                    ambient,	/* ambient color */
                    diffuse,	/* diffuse color */
                    specular;	/* specular color */

    Flt             roughness,	/* specular roughness: 0 < roughn <=1 */
                    index,	/* index of refraction */
                    refl_diffuse,	/* circle of diffusion in reflection */
                    refr_diffuse,	/* circle of diffusion in refraction */
                    brilliance;	/* tightness of diffuse illum. */

    matrix          inv_trans;	/* inverse transformation */
};

struct pixel_data {
    int             imagex,
                    imagey;	/* resolution */
    byte           *pal,
                   *data;	/* the data */
    int             palsize;	/* palette size, 0 for 24 bit color */
};

struct image_map {
    struct pixel_data *pic;	/* the picture */
    void
                    (*inv_map) /* inverse mapping */ (Flt * u, Flt * v, vector
						      x);
    bool            once;	/* only one time? */
    bool            interpolated;	/* not used */
    bool            usenormal,
                    uvswap;	/* swap output of inv_map? */
    Flt             u_offs,
                    v_offs;	/* translate image over surface */
    Flt             u_range,
                    v_range;	/* scale it. */
};

/* eye viewing stuff */
struct camera {
    vector          eye_dir,
                    eye,
                    sky;
    Flt             fov,
                    aspect;
};

/*
 * the object structure.
 */
struct _obj {
    int             type;	/* type of object */
    union shape_data data;	/* pointers to shape data */
    vector         *bbox;	/* for bounding */
    vector         *speed;	/* motion coefficients */
    object         *daddy,	/* for texturing */
                   *next,	/* for link list */
                   *bound,	/* bounding shapes to the object */
                   *clip;	/* clipping shapes */
    long            last_ray_id;/* You have new mail in /usr/object/last_ray_id :) */
    int             flags;	/* miscellaneous info bits */
    struct texture_data *text;	/* the texture of the object */
    struct methods *methods;	/* pointer to shape's procedures */
    matrix         *inv_trans,	/* inverse transformation (if applicable) */
                   *trans;	/* transformation; killed after precomputation. */
};

/* structure for a polynomial in one var. */
typedef struct {
    int             ord;	/* degree of polynomial */
    LFlt            coef[MAX_ORDER + 1];	/* coefficients */
}

spoly;
