/*
 * Shade.c -- the shading function
 *
 * (c) 1994 by Shawn McHorse <Zaphod@fcircus.sat.tx.us> and Han-Wen
 * Nienhuys <hanwen@stack.urc.tue.nl>
 *
 * Based on code by George Kyriazis
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include	<math.h>
#include	"ray.h"
#include	"proto.h"
#include	"extern.h"

PRIVATE color   local_shading(struct intersect *i, struct texture_data *tp, grid_tree * g);

/* global shading model */
PUBLIC color
shade(struct intersect *i, grid_tree * g,
      Flt imp)
{				/* imp is for the depth cutoff */
    color           col,
                    tmpcol;
    object         *op;
    vector          refl;
    Flt             c1 = 0.0,
                    importance,
                    nu,
                    filter_factor,
                    discr;
    struct texture_data t;
    struct ray      reflected,
                    r = i->r;

    /* Find the first object in the hierarchy with a texture. */
    for (op = i->q_ent.obj; op->text == NULL; op = op->daddy);

    /* Find texture at intersection point. */
    fill_texture(&t, op->text, i);
    col = local_shading(i, &t, g);

    /*
     * Don't bother with reflected and refracted rays if we're already as deep as we can go.
     */
    if (g->recursion >= max_trace_level)
	return col;

    reflected.pos = i->x;
    reflected.mint = INFTY;

    importance = imp * color_length(t.refr_color);
    if (importance > depth_cutoff) {
	/* ***************REFRACTION************** */

	ior_depth++;
	if (i->q_ent.entering)
	    global_ior[ior_depth] = t.index;
	else
	    global_ior[ior_depth] = global_ior[ior_depth - 2];

	/* Calculate refracted vector. */
	c1 = -1 * vdot(r.dir, i->n);
	nu = global_ior[ior_depth] / global_ior[ior_depth - 1];
	discr = nu * nu - 1 + c1 * c1;

	/* Total Internal Reflection */
	if (discr < -EPSILON) {

	    /* Calculate reflected vector. */
	    svproduct(refl, 2 * c1, i->n);
	    vadd(refl, r.dir, refl);

	    /* Set up reflected ray. */
	    ior_depth--;
	    reflected.dir = refl;
	    reflected.type = R_REFLECT;
	    reflected.rayid = global_rayid++;

	    /*
	     * we're reflecting and refracting at the same time, so take means
	     */
	    tmpFlt = (t.refl_diffuse + t.refr_diffuse) / 2.0;
	    sample_ray(&reflected, &filter_factor, tmpFlt, g);

	    tmpcol = trace(reflected, g->refract_grid, importance + imp * color_length(t.refl_color));
	    tmpcol = scolorproduct(0.5 * filter_factor, tmpcol);	/* 0.5: reflecting AND refracting */

	    return color_add(col, color_mul(tmpcol, color_add(t.refr_color, t.refl_color)));
	}
	if (discr < EPSILON)
	    svproduct(refl, c1, i->n);
	else
	    svproduct(refl, c1 - sqrt(discr), i->n);
	vadd(refl, r.dir, refl);
	svproduct(refl, 1 / nu, refl);

	/* Set up the refracted ray. */
	reflected.dir = refl;
	reflected.type = R_REFRACT;
	reflected.rayid = global_rayid++;

	sample_ray(&reflected, &filter_factor, t.refr_diffuse, g);


	tmpcol = trace(reflected, g->refract_grid, importance);
	tmpcol = scolorproduct(filter_factor, tmpcol);

	col = color_add(col, color_mul(t.refr_color, tmpcol));
	ior_depth--;
    }
    importance = imp * color_length(t.refl_color);

    if (importance > depth_cutoff) {
	/* ***************REFLECTION************** */

	/* Calculate reflected vector. */
	if (c1 == 0.0)
	    c1 = -1 * vdot(r.dir, i->n);
	svproduct(refl, 2 * c1, i->n);
	vadd(refl, r.dir, refl);

	/* Set up the reflected ray. */
	reflected.dir = refl;
	reflected.type = R_REFLECT;
	reflected.rayid = global_rayid++;
	reflected.mint = INFTY;
	sample_ray(&reflected, &filter_factor, t.refl_diffuse, g);
	tmpcol = trace(reflected, g->reflect_grid, importance);
	tmpcol = scolorproduct(filter_factor, tmpcol);


	col = color_add(col, color_mul(t.refl_color, tmpcol));
    }
    return col;
}

PRIVATE         Flt
                distribution(struct texture_data * tp, vector ldir, vector raydir, vector normal, int type);


/* shading dependent on lightsources */
PRIVATE color
local_shading(struct intersect *i, struct texture_data *tp, grid_tree * g)
{
    object         *op;
    color           col,
                    factor;	/* factor to put influences of distance, brilliance, etc. in */
    int             light_index;
    vector          ldir,
                    normal;
    Flt             ldist,
                    filter_factor,
                    ldotn;
    struct light_data *l;
    struct ray      shadow_ray,
                   *r;

    r = &i->r;

    /* Start off with ambient light. */
    col = color_mul(ambient_light, tp->ambient);

    /* Set up the shadow ray data. */
    shadow_ray.type = R_SHADOW;
    shadow_ray.pos = i->x;

    /* the normal should be looking at us */
    if (vdot(i->n, r->dir) > 0)
	vneg(i->n, i->n);	/* flip it */

    normal = i->n;

    /* Walk the light list. */
    for (light_index = 0; light_index < light_list_len; light_index++) {
	op = light_list[light_index];
	l = op->data.light;

	if (l->type == L_NORMAL) {
	    vector          lorg;
	    ibuf           *dq;

	    if (op->speed) {
		/* incorporate the lightsource's speed into the calculation */
		svproduct(lorg, Time, *op->speed);

	    } else
		setvector(lorg, 0, 0, 0);

	    vadd(lorg, lorg, l->org);

	    /* Construct a normalized vector from object to light source. */
	    vsub(ldir, lorg, i->x);


	    ldist = veclen(ldir);
	    if (ldist < EPSILON)
		continue;
	    svproduct(ldir, 1 / ldist, ldir);

	    shadow_ray.dir = ldir;
	    shadow_ray.rayid = global_rayid++;
	    sample_ray(&shadow_ray, &filter_factor, l->radius / ldist, g);


	    /* Check if the light source is behind the object. */
	    ldotn = vdot(shadow_ray.dir, normal);
	    if (ldotn < EPSILON)
		continue;

	    dq = get_new_ibuf();
	    if (!(op->flags & O_NO_SHADOW) && shadow_test(dq, shadow_ray, ldist, light_index, g)) {
		free_ibuf(dq);
		continue;
	    }
	    free_ibuf(dq);

	} else
	    continue;

	/* Initialize factor. */
	setcolor(factor, 0, 0, 0);

	/* Diffuse shading. */
	if (color_length(tp->diffuse) > EPSILON)
	    if (tp->brilliance != 1.0)
		factor = scolorproduct(pow(ldotn, tp->brilliance), tp->diffuse);
	    else
		factor = scolorproduct(ldotn, tp->diffuse);

	/* Specular shading. */
	if (color_length(tp->specular) > EPSILON && !(op->flags & O_NO_REFLECT)) {

	    /* Apply the local shading model. */
	    factor = color_add(scolorproduct(
				       distribution(tp, shadow_ray.dir, r->dir, normal, tp->distribution), tp->specular), factor);
	}
	if (l->attenuation_exp > EPSILON)
	    factor = scolorproduct(pow(ldist, -1 * l->attenuation_exp), factor);

	col = color_add(col, color_mul(factor, light_color(l, shadow_ray.pos)));
    }

    return col;
}

/*
 * Calculate various microfacet distributions
 */
PRIVATE         Flt
distribution(struct texture_data * tp, vector ldir, vector raydir, vector normal, int type)
{
    Flt             factor;
    vector          viewdir,	/* from surface to eye */
                    refl,	/* reflection */
                    halfway;
    Flt             alpha,
                    exponent,
                    ldotn,
                    hdotn,
                    rdotv;

    norm(viewdir, raydir);
    vneg(viewdir, viewdir);

    factor = 0;
    if (ISZERO(tp->roughness))
	return 0;

    /* phong is the weird duck in the whack :) */
    if (type == PHONG_DISTRIBUTION) {
	ldotn = vdot(ldir, normal);

	/* Calculate the reflected vector. */
	svproduct(refl, 2 * ldotn, normal);
	vneg(tmpvector, ldir);
	vadd(refl, tmpvector, refl);

	rdotv = vdot(refl, viewdir);

	/* Initialize factor. */
	if (rdotv > EPSILON)
	    if (tp->roughness != 1.0)
		factor = pow(rdotv, 1 / tp->roughness);
	    else
		factor = rdotv;

	return factor;
    }
    vadd(halfway, ldir, viewdir);
    norm(halfway, halfway);
    hdotn = vdot(halfway, normal);
    if (hdotn < EPSILON)
	return 0;

    switch (type) {
    case BLINN_DISTRIBUTION:
	if (tp->roughness != 1.0)
	    factor = pow(hdotn, 1 / tp->roughness);
	else
	    factor = hdotn;
	break;

    case GAUSS_DISTRIBUTION:
	alpha = acos(hdotn);
	exponent = (alpha / tp->roughness);
	exponent = -exponent * exponent;
	factor = exp(exponent);
	break;

    case COOK_DISTRIBUTION:
	{
	    Flt             c,
	                    r;

	    c = sqr(hdotn);
	    r = sqr(tp->roughness);
	    exponent = -((1 / c) - 1) / r;
	    c = c * c;
	    factor = exp(exponent) / (c * 4 * M_PI * r);
	}
	break;

    case REITZ_DISTRIBUTION:
	{
	    Flt             c,
	                    r;

	    c = sqr(hdotn);
	    r = sqr(tp->roughness);

	    /*
	     * there is an error in the book "Image Synthesis" by Thalmann & Magenat-Thalman
	     */
	    factor = r / (c * (r - 1) + 1);
	    factor = sqr(factor);

	}
	break;
    default:
	assert(FALSE);
    }

    return factor;
}

/*
 * Stuff below is based on the sources by Roy Hall:
 *
 * "Illumination and Color in Computer Generated Imagery" Source for
 * code given in the Appendices (c)Roy Hall - 1988,1989
 *
 *
 * I am making this code available to the user community.  You are
 * free to use it as you wish.  In return I request two things.
 * First, that this README file always stays with any copy of the
 * source, more specifically, that proper credit is given to the
 * original author. Second, that any corrections, improvements,
 * etc. are passed back to the author for inclusion in subsequent
 * distribution and editions of the book.

   I copied a small part of the D.c file, and adapted it to my shading
   code. Including the whole README seemed a bit of overkill
*/

/* find roughness based on the angle at which the intensity is half of the max */
PUBLIC          Flt
find_roughness(Flt beta, int type)
{
    if (beta < EPSILON || beta > M_PI / 2)
	return 0;

    switch (type) {
    case PHONG_DISTRIBUTION:
	{
	    if (beta >= M_PI_4)
		return 0.0;
	    return 1 / (-(log(2.0) / log(cos(2.0 * beta))));
	}
	break;

    case BLINN_DISTRIBUTION:

	{
	    return 1 / (-(log(2.0) / log(cos(beta))));
	}
	break;

    case GAUSS_DISTRIBUTION:
	{
	    return 1 / (sqrt(log(2.0)) / beta);
	}
	break;

    case COOK_DISTRIBUTION:
	{
	    Flt             tan_beta;

	    tan_beta = tan(beta);
	    return sqrt(-(tan_beta * tan_beta) /
			log(pow(cos(beta), 4.0) / 2.0));
	}
	break;

    case REITZ_DISTRIBUTION:
	{
	    Flt             cos_beta;

	    cos_beta = cos(beta);
	    return sqrt(((cos_beta * cos_beta) - 1.0) /
			((cos_beta * cos_beta) - sqrt(2.0)));
	}
	break;
    default:
	assert(FALSE);
    }
}

/*eof*/
