/*
 * sphere.c -- standard sphere routines.
 *
 * (c) 1993, 1994 by Han-Wen Nienhuys <hanwen@stack.urc.tue.nl>
 *
 * This is based on work by George Kyriazis.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "ray.h"
#include "proto.h"
#include "extern.h"

extern struct methods my_methods;


PRIVATE bool
all_sphere_intersections(ibuf * q, object *o, struct ray *r,
			 int flags,	/* give all intersections */
			 bool *isinside)
{				/* for use with CSG: isinside == inside_sphere() */
    dqueue          q_ent;
    int             no_int;
    Flt             d[2];

    no_int = intersect_sphere(o->data.sphere->center, o->data.sphere->rsq, r, d);

    *isinside = (no_int % 2) ^ (!!(o->flags & O_INVERTED));

    if (!no_int)
	return FALSE;


    q_ent.t = d[0];
    q_ent.obj = o;
    q_ent.entering = !(*isinside);
    add_to_ibuf(q, flags, &q_ent);

    if ((flags & CHKALL) && no_int == 2) {
	q_ent.t = d[1];
	q_ent.entering = (*isinside);
	add_to_ibuf(q, flags, &q_ent);
    }
    return TRUE;
}

/* Intersect ray with sphere */
PUBLIC int
intersect_sphere(vector center, Flt rsq, struct ray *r, Flt * hits)
{
    vector          center_direction;
    Flt             discriminant,
                    linear,
                    dir_projection,
                    constant,
                    smallzero,
                    quadratic,
                    cd_lensq;
    Flt             solmax,
                    solmin;


    my_methods.test++;


    /*
     * solve: x = (pos+t*dir) and |x-center|^2  = radius^2
     */
    vsub(center_direction, r->pos, center);	/* vector from center to r.pos */
    dir_projection = vdot(r->dir, center_direction);
    cd_lensq = vdot(center_direction, center_direction);	/* square of length of center_direction */

    if (cd_lensq >= rsq &&	/* we're not inside the sphere */
	dir_projection > 0) {	/* and we're not looking at the sphere */
	return 0;
    }
    linear = 2 * dir_projection;
    constant = cd_lensq - rsq;

    /* with the advent of extrusions, |r.dir| isn't always 1 */
    quadratic = vdot(r->dir, r->dir);
    smallzero = (constant / linear);

    /* avoid nasty roots at zero */
    if (smallzero < tolerance && smallzero > -tolerance) {
	solmin = -linear / quadratic;

	if (solmin > tolerance) {
	    *hits = solmin;
	    return 1;
	} else
	    return 0;
    }
    discriminant = sqr(linear) - 4 * quadratic * constant;

    if (discriminant < 0)
	return 0;

    quadratic *= 2;
    discriminant = sqrt(discriminant);
    solmax = (-linear + discriminant) / (quadratic);
    solmin = (-linear - discriminant) / (quadratic);

    if (solmax < tolerance)
	return 0;		/* maximum is  behind us. Shake it. */

    /* This is an intersection! */
    my_methods.hit++;
    if (solmin < tolerance) {
	*hits = solmax;
	return 1;
    } else {
	*hits++ = solmin;
	*hits = solmax;
	return 2;
    }

}

PRIVATE vector
sphere_normal(struct intersect *i, vector x)
{
    vector          n;
    object         *o = i->q_ent.obj;

    /* calculate the normal.  It is just the direction of the radius */
    /* norm(i.x - cntr) */
    vsub(n, x, o->data.sphere->center);

    return n;
}

PRIVATE void
precompute_sphere(object *o)
{
    struct sphere_data *sph = o->data.sphere;

    sph->rsq = sqr(sph->radius);
    sph->radius = ABS(sph->radius);

    /* bounding volume calculation, GGems */

    /*
     * The bounding volume of a sphere is found.  The algorithm
     * performs three passes, one for each dimension.  In each pass
     * the extrema of the surface are found as values of the
     * parameters u and v.  These values are then used to calculate
     * the position of the two points in the current
     * dimension. Finally, these points are sorted to find the minimum
     * and maximum extents. The canonical sphere has a radius of 1.0
     * and is centered at the origin.
     */


    {
	Flt             min[3],
	                max[3];

	/* returned minimum and maximum of extent */

	matrix          ctm,	/* cumulative transformation matrix */
	                tempmatrix;
	int             i;
	float           u1,
	                u2,
	                v1,
	                v2,
	                tmp,
	                denominator;
	vector          scal;

	setvector(scal, sph->radius, sph->radius, sph->radius);
	scale_matrix(ctm, scal);
	translate_matrix(tempmatrix, sph->center);
	mmproduct(tempmatrix, tempmatrix, ctm);

	if (o->inv_trans) {
	    mmproduct(tempmatrix, *o->trans, tempmatrix);
	}
	transpose_matrix(ctm, tempmatrix);

	for (i = 0; i < 3; i++) {

	    /* calculate first extremum. */
	    u1 = safe_arctangent(ctm[2][i], ctm[0][i]);
	    denominator = ctm[0][i] * cos(u1) + ctm[2][i] * sin(u1);
	    v1 = safe_arctangent(ctm[1][i], denominator);

	    /* second extremum is +/- PI from u1, negative of v1 */
	    if (u1 <= 0)
		u2 = u1 + M_PI;
	    else
		u2 = u1 - M_PI;
	    v2 = -v1;

	    /* find and sort extrema locations in this dimension */
	    min[i] =
		ctm[0][i] * cos(u1) * cos(v1) +
		ctm[1][i] * sin(v1) +
		ctm[2][i] * sin(u1) * cos(v1) +
		ctm[3][i];
	    max[i] =
		ctm[0][i] * cos(u2) * cos(v2) +
		ctm[1][i] * sin(v2) +
		ctm[2][i] * sin(u2) * cos(v2) +
		ctm[3][i];
	    if (min[i] > max[i]) {
		tmp = max[i];
		max[i] = min[i];
		min[i] = tmp;
	    }
	}
	o->bbox[0].x = min[0];
	o->bbox[0].y = min[1];
	o->bbox[0].z = min[2];
	o->bbox[1].x = max[0];
	o->bbox[1].y = max[1];
	o->bbox[1].z = max[2];
    }

    my_methods.howmuch++;
    global_stats.prims++;
}

PRIVATE bool
inside_sphere(object *o, vector x)
{
    vsub(x, x, o->data.sphere->center);
    if (vdot(x, x) < o->data.sphere->rsq)
	return TRUE;
    else
	return FALSE;
}

PRIVATE void
translate_sphere(object *o, vector t)
{
    struct sphere_data *s;

    s = o->data.sphere;
    if (o->inv_trans)
	generic_translate_object(o, t);
    else
	vadd(s->center, s->center, t);
}

PRIVATE void
scale_sphere(object *o, vector s)
{
    struct sphere_data *sph;
    Flt             factor;

    sph = o->data.sphere;

    if (o->inv_trans || ABS(s.x) != ABS(s.y) || ABS(s.y) != ABS(s.z)) {
	/* scale unevenly */
	generic_scale_object(o, s);
    } else {
	factor = ABS(s.x);
	sph->radius *= factor;
	vcproduct(sph->center, s, sph->center);
    }
}

PRIVATE void
rotate_sphere(object *o, matrix m)
{
    struct sphere_data *sph;

    sph = o->data.sphere;
    if (o->inv_trans != NULL)
	generic_rotate_object(o, m);
    else
	rotate_vector(&sph->center, m);
}

PRIVATE void
init_sphere(struct sphere_data *s)
{
    setvector(s->center, 0, 0, 0);
    s->radius = 1.0;
}

PRIVATE struct sphere_data *
get_new_sphere(void)
{
    struct sphere_data *s;

    s = ALLOC(struct sphere_data);

    init_sphere(s);
    return s;
}

PRIVATE void
copy_sphere(object *dst, object *src)
{
    assert(dst != NULL && src != NULL);
    if (dst->type != SPHERE)
	dst->data.sphere = get_new_sphere();
    *dst->data.sphere = *src->data.sphere;
    dst->type = src->type;
}

PRIVATE void
free_sphere(object *s)
{
    free((void *) s->data.sphere);
    s->type = NOSHAPE;
}

PRIVATE void
print_sphere(object *o)
{
#ifdef DEBUG
    struct sphere_data *s = o->data.sphere;

    dprintf("radius %f ", s->radius);
    print_v("cntr", s->center);
#endif				/* DEBUG */
}


PRIVATE struct methods my_methods =
{
    all_sphere_intersections,
    sphere_normal,
    copy_sphere,
    inside_sphere,
    rotate_sphere,
    translate_sphere,
    scale_sphere,
    free_sphere,
    print_sphere,
    precompute_sphere,
    "sphere",
};


PUBLIC object  *
get_new_sphere_object(void)
{
    object         *o;

    o = get_new_object();
    o->data.sphere = get_new_sphere();
    o->methods = &my_methods;
    o->type = SPHERE;
    o->flags = O_CLOSED | O_FINITE_INSIDE;

    return o;
}
